from typing import Dict, Union

from ..enums.ClassIDType import ClassIDType as CIT
from . import (
    AimConstraint,
    AnchoredJoint2D,
    Animation,
    AnimationClip,
    Animator,
    AnimatorController,
    AnimatorOverrideController,
    AnimatorState,
    AnimatorStateMachine,
    AnimatorStateTransition,
    AnimatorTransition,
    AnimatorTransitionBase,
    AnnotationManager,
    AreaEffector2D,
    ArticulationBody,
    AssemblyDefinitionAsset,
    AssemblyDefinitionImporter,
    AssemblyDefinitionReferenceAsset,
    AssemblyDefinitionReferenceImporter,
    AssetBundle,
    AssetBundleManifest,
    AssetDatabaseV1,
    AssetImporter,
    AssetImporterLog,
    AssetImportInProgressProxy,
    AssetMetaData,
    AssetServerCache,
    ASTCImporter,
    AudioBehaviour,
    AudioBuildInfo,
    AudioChorusFilter,
    AudioClip,
    AudioDistortionFilter,
    AudioEchoFilter,
    AudioFilter,
    AudioHighPassFilter,
    AudioImporter,
    AudioListener,
    AudioLowPassFilter,
    AudioManager,
    AudioMixer,
    AudioMixerController,
    AudioMixerEffectController,
    AudioMixerGroup,
    AudioMixerGroupController,
    AudioMixerLiveUpdateBool,
    AudioMixerLiveUpdateFloat,
    AudioMixerSnapshot,
    AudioMixerSnapshotController,
    AudioReverbFilter,
    AudioReverbZone,
    AudioSource,
    Avatar,
    AvatarMask,
    BaseAnimationTrack,
    BaseVideoTexture,
    Behaviour,
    BillboardAsset,
    BillboardRenderer,
    BlendTree,
    BoxCollider,
    BoxCollider2D,
    BuildReport,
    BuildSettings,
    BuiltAssetBundleInfoSet,
    BuoyancyEffector2D,
    CachedSpriteAtlas,
    CachedSpriteAtlasRuntimeData,
    Camera,
    Canvas,
    CanvasGroup,
    CanvasRenderer,
    CapsuleCollider,
    CapsuleCollider2D,
    CGProgram,
    CharacterController,
    CharacterJoint,
    CircleCollider2D,
    Cloth,
    ClothRenderer,
    CloudWebServicesManager,
    ClusterInputManager,
    Collider,
    Collider2D,
    Collision,
    Collision2D,
    Component,
    CompositeCollider2D,
    ComputeShader,
    ComputeShaderImporter,
    ConfigurableJoint,
    ConstantForce,
    ConstantForce2D,
    CrashReportManager,
    Cubemap,
    CubemapArray,
    CustomRenderTexture,
    DDSImporter,
    DefaultAsset,
    DefaultImporter,
    DelayedCallManager,
    Derived,
    DistanceJoint2D,
    EdgeCollider2D,
    EditorBuildSettings,
    EditorExtension,
    EditorExtensionImpl,
    EditorProjectAccess,
    EditorSettings,
    EditorUserBuildSettings,
    EditorUserSettings,
    Effector2D,
    EllipsoidParticleEmitter,
    EmptyObject,
    FakeComponent,
    FBXImporter,
    FixedJoint,
    FixedJoint2D,
    Flare,
    FlareLayer,
    Font,
    FrictionJoint2D,
    GameManager,
    GameObject,
    GameObjectRecorder,
    GlobalGameManager,
    GraphicsSettings,
    Grid,
    GridLayout,
    GUIDSerializer,
    GUIElement,
    GUILayer,
    GUIText,
    GUITexture,
    Halo,
    HaloLayer,
    HaloManager,
    HierarchyState,
    HingeJoint,
    HingeJoint2D,
    HumanTemplate,
    IConstraint,
    IHVImageFormatImporter,
    InputManager,
    InspectorExpandedState,
    InteractiveCloth,
    Joint,
    Joint2D,
    KTXImporter,
    LensFlare,
    LevelGameManager,
    LibraryAssetImporter,
    Light,
    LightingDataAsset,
    LightingDataAssetParent,
    LightingSettings,
    LightmapParameters,
    LightmapSettings,
    LightProbeGroup,
    LightProbeProxyVolume,
    LightProbes,
    LineRenderer,
    LocalizationAsset,
    LocalizationImporter,
    LODGroup,
    LookAtConstraint,
    LowerResBlitTexture,
    MasterServerInterface,
    Material,
    Mesh,
    Mesh3DSImporter,
    MeshCollider,
    MeshFilter,
    MeshParticleEmitter,
    MeshRenderer,
    ModelImporter,
    MonoBehaviour,
    MonoImporter,
    MonoManager,
    MonoObject,
    MonoScript,
    Motion,
    MovieImporter,
    MovieTexture,
    MultiArtifactTestImporter,
    NamedObject,
    NativeFormatImporter,
    NativeObjectType,
    NavMeshAgent,
    NavMeshData,
    NavMeshObsolete,
    NavMeshObstacle,
    NavMeshProjectSettings,
    NavMeshSettings,
    NetworkManager,
    NetworkView,
    NewAnimationTrack,
    NScreenBridge,
    Object,
    OcclusionArea,
    OcclusionCullingData,
    OcclusionCullingSettings,
    OcclusionPortal,
    OffMeshLink,
    PackageManifest,
    PackageManifestImporter,
    PackedAssets,
    ParentConstraint,
    ParticleAnimator,
    ParticleEmitter,
    ParticleRenderer,
    ParticleSystem,
    ParticleSystemForceField,
    ParticleSystemRenderer,
    PerformanceReportingManager,
    PhysicMaterial,
    Physics2DSettings,
    PhysicsManager,
    PhysicsMaterial2D,
    PhysicsUpdateBehaviour2D,
    Pipeline,
    PlatformEffector2D,
    PlatformModuleSetup,
    PlayableDirector,
    PlayerSettings,
    PluginBuildInfo,
    PluginImporter,
    PointEffector2D,
    Polygon2D,
    PolygonCollider2D,
    PositionConstraint,
    PrefabImporter,
    PrefabInstance,
    PreloadData,
    Preset,
    PresetManager,
    PreviewAnimationClip,
    ProceduralMaterial,
    ProceduralTexture,
    Projector,
    PropertyModificationsTargetTestObject,
    PVRImporter,
    QualitySettings,
    RaycastCollider,
    RayTracingShader,
    RayTracingShaderImporter,
    RectTransform,
    ReferencesArtifactGenerator,
    ReflectionProbe,
    RelativeJoint2D,
    Renderer,
    RendererFake,
    RenderPassAttachment,
    RenderSettings,
    RenderTexture,
    ResourceManager,
    Rigidbody,
    Rigidbody2D,
    RootMotionData,
    RotationConstraint,
    RuntimeAnimatorController,
    RuntimeInitializeOnLoadManager,
    SampleClip,
    ScaleConstraint,
    SceneAsset,
    ScenesUsingAssets,
    SceneVisibilityState,
    ScriptableCamera,
    ScriptedImporter,
    ScriptMapper,
    SerializableManagedHost,
    SerializableManagedRefTestClass,
    Shader,
    ShaderImporter,
    ShaderVariantCollection,
    SiblingDerived,
    SketchUpImporter,
    SkinnedCloth,
    SkinnedMeshRenderer,
    Skybox,
    SliderJoint2D,
    SortingGroup,
    SparseTexture,
    SpeedTreeImporter,
    SpeedTreeWindAsset,
    SphereCollider,
    SpringJoint,
    SpringJoint2D,
    Sprite,
    SpriteAtlas,
    SpriteAtlasAsset,
    SpriteAtlasDatabase,
    SpriteAtlasImporter,
    SpriteMask,
    SpriteRenderer,
    SpriteShapeRenderer,
    StreamingController,
    StreamingManager,
    SubDerived,
    SubstanceArchive,
    SubstanceImporter,
    SurfaceEffector2D,
    TagManager,
    TargetJoint2D,
    Terrain,
    TerrainCollider,
    TerrainData,
    TerrainLayer,
    TestObjectVectorPairStringBool,
    TestObjectWithSerializedAnimationCurve,
    TestObjectWithSerializedArray,
    TestObjectWithSerializedMapStringBool,
    TestObjectWithSerializedMapStringNonAlignedStruct,
    TestObjectWithSpecialLayoutOne,
    TestObjectWithSpecialLayoutTwo,
    TextAsset,
    TextMesh,
    TextScriptImporter,
    Texture,
    Texture2D,
    Texture2DArray,
    Texture3D,
    TextureImporter,
    Tilemap,
    TilemapCollider2D,
    TilemapRenderer,
    TimeManager,
    TrailRenderer,
    Transform,
    Tree,
    TrueTypeFontImporter,
    UnityAdsManager,
    UnityAnalyticsManager,
    UnityConnectSettings,
    Vector3f,
    VersionControlSettings,
    VFXManager,
    VFXRenderer,
    VideoClip,
    VideoClipImporter,
    VideoPlayer,
    VisualEffect,
    VisualEffectAsset,
    VisualEffectImporter,
    VisualEffectObject,
    VisualEffectResource,
    VisualEffectSubgraph,
    VisualEffectSubgraphBlock,
    VisualEffectSubgraphOperator,
    WebCamTexture,
    WheelCollider,
    WheelJoint2D,
    WindZone,
    WorldAnchor,
    WorldParticleCollider,
)

ClassIDTypeToClassMapValueType = Union[
    AimConstraint,
    AnchoredJoint2D,
    Animation,
    AnimationClip,
    Animator,
    AnimatorController,
    AnimatorOverrideController,
    AnimatorState,
    AnimatorStateMachine,
    AnimatorStateTransition,
    AnimatorTransition,
    AnimatorTransitionBase,
    AnnotationManager,
    AreaEffector2D,
    ArticulationBody,
    AssemblyDefinitionAsset,
    AssemblyDefinitionImporter,
    AssemblyDefinitionReferenceAsset,
    AssemblyDefinitionReferenceImporter,
    AssetBundle,
    AssetBundleManifest,
    AssetDatabaseV1,
    AssetImporter,
    AssetImporterLog,
    AssetImportInProgressProxy,
    AssetMetaData,
    AssetServerCache,
    ASTCImporter,
    AudioBehaviour,
    AudioBuildInfo,
    AudioChorusFilter,
    AudioClip,
    AudioDistortionFilter,
    AudioEchoFilter,
    AudioFilter,
    AudioHighPassFilter,
    AudioImporter,
    AudioListener,
    AudioLowPassFilter,
    AudioManager,
    AudioMixer,
    AudioMixerController,
    AudioMixerEffectController,
    AudioMixerGroup,
    AudioMixerGroupController,
    AudioMixerLiveUpdateBool,
    AudioMixerLiveUpdateFloat,
    AudioMixerSnapshot,
    AudioMixerSnapshotController,
    AudioReverbFilter,
    AudioReverbZone,
    AudioSource,
    Avatar,
    AvatarMask,
    BaseAnimationTrack,
    BaseVideoTexture,
    Behaviour,
    BillboardAsset,
    BillboardRenderer,
    BlendTree,
    BoxCollider,
    BoxCollider2D,
    BuildReport,
    BuildSettings,
    BuiltAssetBundleInfoSet,
    BuoyancyEffector2D,
    CachedSpriteAtlas,
    CachedSpriteAtlasRuntimeData,
    Camera,
    Canvas,
    CanvasGroup,
    CanvasRenderer,
    CapsuleCollider,
    CapsuleCollider2D,
    CGProgram,
    CharacterController,
    CharacterJoint,
    CircleCollider2D,
    Cloth,
    ClothRenderer,
    CloudWebServicesManager,
    ClusterInputManager,
    Collider,
    Collider2D,
    Collision,
    Collision2D,
    Component,
    CompositeCollider2D,
    ComputeShader,
    ComputeShaderImporter,
    ConfigurableJoint,
    ConstantForce,
    ConstantForce2D,
    CrashReportManager,
    Cubemap,
    CubemapArray,
    CustomRenderTexture,
    DDSImporter,
    DefaultAsset,
    DefaultImporter,
    DelayedCallManager,
    Derived,
    DistanceJoint2D,
    EdgeCollider2D,
    EditorBuildSettings,
    EditorExtension,
    EditorExtensionImpl,
    EditorProjectAccess,
    EditorSettings,
    EditorUserBuildSettings,
    EditorUserSettings,
    Effector2D,
    EllipsoidParticleEmitter,
    EmptyObject,
    FakeComponent,
    FBXImporter,
    FixedJoint,
    FixedJoint2D,
    Flare,
    FlareLayer,
    Font,
    FrictionJoint2D,
    GameManager,
    GameObject,
    GameObjectRecorder,
    GlobalGameManager,
    GraphicsSettings,
    Grid,
    GridLayout,
    GUIDSerializer,
    GUIElement,
    GUILayer,
    GUIText,
    GUITexture,
    Halo,
    HaloLayer,
    HaloManager,
    HierarchyState,
    HingeJoint,
    HingeJoint2D,
    HumanTemplate,
    IConstraint,
    IHVImageFormatImporter,
    InputManager,
    InspectorExpandedState,
    InteractiveCloth,
    Joint,
    Joint2D,
    KTXImporter,
    LensFlare,
    LevelGameManager,
    LibraryAssetImporter,
    Light,
    LightingDataAsset,
    LightingDataAssetParent,
    LightingSettings,
    LightmapParameters,
    LightmapSettings,
    LightProbeGroup,
    LightProbeProxyVolume,
    LightProbes,
    LineRenderer,
    LocalizationAsset,
    LocalizationImporter,
    LODGroup,
    LookAtConstraint,
    LowerResBlitTexture,
    MasterServerInterface,
    Material,
    Mesh,
    Mesh3DSImporter,
    MeshCollider,
    MeshFilter,
    MeshParticleEmitter,
    MeshRenderer,
    ModelImporter,
    MonoBehaviour,
    MonoImporter,
    MonoManager,
    MonoObject,
    MonoScript,
    Motion,
    MovieImporter,
    MovieTexture,
    MultiArtifactTestImporter,
    NamedObject,
    NativeFormatImporter,
    NativeObjectType,
    NavMeshAgent,
    NavMeshData,
    NavMeshObsolete,
    NavMeshObstacle,
    NavMeshProjectSettings,
    NavMeshSettings,
    NetworkManager,
    NetworkView,
    NewAnimationTrack,
    NScreenBridge,
    Object,
    OcclusionArea,
    OcclusionCullingData,
    OcclusionCullingSettings,
    OcclusionPortal,
    OffMeshLink,
    PackageManifest,
    PackageManifestImporter,
    PackedAssets,
    ParentConstraint,
    ParticleAnimator,
    ParticleEmitter,
    ParticleRenderer,
    ParticleSystem,
    ParticleSystemForceField,
    ParticleSystemRenderer,
    PerformanceReportingManager,
    PhysicMaterial,
    Physics2DSettings,
    PhysicsManager,
    PhysicsMaterial2D,
    PhysicsUpdateBehaviour2D,
    Pipeline,
    PlatformEffector2D,
    PlatformModuleSetup,
    PlayableDirector,
    PlayerSettings,
    PluginBuildInfo,
    PluginImporter,
    PointEffector2D,
    Polygon2D,
    PolygonCollider2D,
    PositionConstraint,
    PrefabImporter,
    PrefabInstance,
    PreloadData,
    Preset,
    PresetManager,
    PreviewAnimationClip,
    ProceduralMaterial,
    ProceduralTexture,
    Projector,
    PropertyModificationsTargetTestObject,
    PVRImporter,
    QualitySettings,
    RaycastCollider,
    RayTracingShader,
    RayTracingShaderImporter,
    RectTransform,
    ReferencesArtifactGenerator,
    ReflectionProbe,
    RelativeJoint2D,
    Renderer,
    RendererFake,
    RenderPassAttachment,
    RenderSettings,
    RenderTexture,
    ResourceManager,
    Rigidbody,
    Rigidbody2D,
    RootMotionData,
    RotationConstraint,
    RuntimeAnimatorController,
    RuntimeInitializeOnLoadManager,
    SampleClip,
    ScaleConstraint,
    SceneAsset,
    ScenesUsingAssets,
    SceneVisibilityState,
    ScriptableCamera,
    ScriptedImporter,
    ScriptMapper,
    SerializableManagedHost,
    SerializableManagedRefTestClass,
    Shader,
    ShaderImporter,
    ShaderVariantCollection,
    SiblingDerived,
    SketchUpImporter,
    SkinnedCloth,
    SkinnedMeshRenderer,
    Skybox,
    SliderJoint2D,
    SortingGroup,
    SparseTexture,
    SpeedTreeImporter,
    SpeedTreeWindAsset,
    SphereCollider,
    SpringJoint,
    SpringJoint2D,
    Sprite,
    SpriteAtlas,
    SpriteAtlasAsset,
    SpriteAtlasDatabase,
    SpriteAtlasImporter,
    SpriteMask,
    SpriteRenderer,
    SpriteShapeRenderer,
    StreamingController,
    StreamingManager,
    SubDerived,
    SubstanceArchive,
    SubstanceImporter,
    SurfaceEffector2D,
    TagManager,
    TargetJoint2D,
    Terrain,
    TerrainCollider,
    TerrainData,
    TerrainLayer,
    TestObjectVectorPairStringBool,
    TestObjectWithSerializedAnimationCurve,
    TestObjectWithSerializedArray,
    TestObjectWithSerializedMapStringBool,
    TestObjectWithSerializedMapStringNonAlignedStruct,
    TestObjectWithSpecialLayoutOne,
    TestObjectWithSpecialLayoutTwo,
    TextAsset,
    TextMesh,
    TextScriptImporter,
    Texture,
    Texture2D,
    Texture2DArray,
    Texture3D,
    TextureImporter,
    Tilemap,
    TilemapCollider2D,
    TilemapRenderer,
    TimeManager,
    TrailRenderer,
    Transform,
    Tree,
    TrueTypeFontImporter,
    UnityAdsManager,
    UnityAnalyticsManager,
    UnityConnectSettings,
    Vector3f,
    VersionControlSettings,
    VFXManager,
    VFXRenderer,
    VideoClip,
    VideoClipImporter,
    VideoPlayer,
    VisualEffect,
    VisualEffectAsset,
    VisualEffectImporter,
    VisualEffectObject,
    VisualEffectResource,
    VisualEffectSubgraph,
    VisualEffectSubgraphBlock,
    VisualEffectSubgraphOperator,
    WebCamTexture,
    WheelCollider,
    WheelJoint2D,
    WindZone,
    WorldAnchor,
    WorldParticleCollider,
]

ClassIDTypeToClassMap: Dict[CIT, ClassIDTypeToClassMapValueType] = {
    CIT.Object: Object,
    CIT.GameObject: GameObject,
    CIT.Component: Component,
    CIT.LevelGameManager: LevelGameManager,
    CIT.Transform: Transform,
    CIT.TimeManager: TimeManager,
    CIT.GlobalGameManager: GlobalGameManager,
    CIT.Behaviour: Behaviour,
    CIT.GameManager: GameManager,
    CIT.AudioManager: AudioManager,
    CIT.ParticleAnimator: ParticleAnimator,
    CIT.InputManager: InputManager,
    CIT.EllipsoidParticleEmitter: EllipsoidParticleEmitter,
    CIT.Pipeline: Pipeline,
    CIT.EditorExtension: EditorExtension,
    CIT.Physics2DSettings: Physics2DSettings,
    CIT.Camera: Camera,
    CIT.Material: Material,
    CIT.MeshRenderer: MeshRenderer,
    CIT.Renderer: Renderer,
    CIT.ParticleRenderer: ParticleRenderer,
    CIT.Texture: Texture,
    CIT.Texture2D: Texture2D,
    CIT.OcclusionCullingSettings: OcclusionCullingSettings,
    CIT.GraphicsSettings: GraphicsSettings,
    CIT.MeshFilter: MeshFilter,
    CIT.OcclusionPortal: OcclusionPortal,
    CIT.Mesh: Mesh,
    CIT.Skybox: Skybox,
    CIT.QualitySettings: QualitySettings,
    CIT.Shader: Shader,
    CIT.TextAsset: TextAsset,
    CIT.Rigidbody2D: Rigidbody2D,
    CIT.Physics2DManager: Physics2DSettings,
    CIT.Collider2D: Collider2D,
    CIT.Rigidbody: Rigidbody,
    CIT.PhysicsManager: PhysicsManager,
    CIT.Collider: Collider,
    CIT.Joint: Joint,
    CIT.CircleCollider2D: CircleCollider2D,
    CIT.HingeJoint: HingeJoint,
    CIT.PolygonCollider2D: PolygonCollider2D,
    CIT.BoxCollider2D: BoxCollider2D,
    CIT.PhysicsMaterial2D: PhysicsMaterial2D,
    CIT.MeshCollider: MeshCollider,
    CIT.BoxCollider: BoxCollider,
    CIT.CompositeCollider2D: CompositeCollider2D,
    CIT.EdgeCollider2D: EdgeCollider2D,
    CIT.CapsuleCollider2D: CapsuleCollider2D,
    CIT.ComputeShader: ComputeShader,
    CIT.AnimationClip: AnimationClip,
    CIT.ConstantForce: ConstantForce,
    CIT.WorldParticleCollider: WorldParticleCollider,
    CIT.TagManager: TagManager,
    CIT.AudioListener: AudioListener,
    CIT.AudioSource: AudioSource,
    CIT.AudioClip: AudioClip,
    CIT.RenderTexture: RenderTexture,
    CIT.CustomRenderTexture: CustomRenderTexture,
    CIT.MeshParticleEmitter: MeshParticleEmitter,
    CIT.ParticleEmitter: ParticleEmitter,
    CIT.Cubemap: Cubemap,
    CIT.Avatar: Avatar,
    CIT.AnimatorController: AnimatorController,
    CIT.GUILayer: GUILayer,
    CIT.RuntimeAnimatorController: RuntimeAnimatorController,
    CIT.ScriptMapper: ScriptMapper,
    CIT.Animator: Animator,
    CIT.TrailRenderer: TrailRenderer,
    CIT.DelayedCallManager: DelayedCallManager,
    CIT.TextMesh: TextMesh,
    CIT.RenderSettings: RenderSettings,
    CIT.Light: Light,
    CIT.CGProgram: CGProgram,
    CIT.BaseAnimationTrack: BaseAnimationTrack,
    CIT.Animation: Animation,
    CIT.MonoBehaviour: MonoBehaviour,
    CIT.MonoScript: MonoScript,
    CIT.MonoManager: MonoManager,
    CIT.Texture3D: Texture3D,
    CIT.NewAnimationTrack: NewAnimationTrack,
    CIT.Projector: Projector,
    CIT.LineRenderer: LineRenderer,
    CIT.Flare: Flare,
    CIT.Halo: Halo,
    CIT.LensFlare: LensFlare,
    CIT.FlareLayer: FlareLayer,
    CIT.HaloLayer: HaloLayer,
    CIT.NavMeshProjectSettings: NavMeshProjectSettings,
    CIT.HaloManager: HaloManager,
    CIT.Font: Font,
    CIT.PlayerSettings: PlayerSettings,
    CIT.NamedObject: NamedObject,
    CIT.GUITexture: GUITexture,
    CIT.GUIText: GUIText,
    CIT.GUIElement: GUIElement,
    CIT.PhysicMaterial: PhysicMaterial,
    CIT.SphereCollider: SphereCollider,
    CIT.CapsuleCollider: CapsuleCollider,
    CIT.SkinnedMeshRenderer: SkinnedMeshRenderer,
    CIT.FixedJoint: FixedJoint,
    CIT.RaycastCollider: RaycastCollider,
    CIT.BuildSettings: BuildSettings,
    CIT.AssetBundle: AssetBundle,
    CIT.CharacterController: CharacterController,
    CIT.CharacterJoint: CharacterJoint,
    CIT.SpringJoint: SpringJoint,
    CIT.WheelCollider: WheelCollider,
    CIT.ResourceManager: ResourceManager,
    CIT.NetworkView: NetworkView,
    CIT.NetworkManager: NetworkManager,
    CIT.PreloadData: PreloadData,
    CIT.MovieTexture: MovieTexture,
    CIT.ConfigurableJoint: ConfigurableJoint,
    CIT.TerrainCollider: TerrainCollider,
    CIT.MasterServerInterface: MasterServerInterface,
    CIT.TerrainData: TerrainData,
    CIT.LightmapSettings: LightmapSettings,
    CIT.WebCamTexture: WebCamTexture,
    CIT.EditorSettings: EditorSettings,
    CIT.InteractiveCloth: InteractiveCloth,
    CIT.ClothRenderer: ClothRenderer,
    CIT.EditorUserSettings: EditorUserSettings,
    CIT.SkinnedCloth: SkinnedCloth,
    CIT.AudioReverbFilter: AudioReverbFilter,
    CIT.AudioHighPassFilter: AudioHighPassFilter,
    CIT.AudioChorusFilter: AudioChorusFilter,
    CIT.AudioReverbZone: AudioReverbZone,
    CIT.AudioEchoFilter: AudioEchoFilter,
    CIT.AudioLowPassFilter: AudioLowPassFilter,
    CIT.AudioDistortionFilter: AudioDistortionFilter,
    CIT.SparseTexture: SparseTexture,
    CIT.AudioBehaviour: AudioBehaviour,
    CIT.AudioFilter: AudioFilter,
    CIT.WindZone: WindZone,
    CIT.Cloth: Cloth,
    CIT.SubstanceArchive: SubstanceArchive,
    CIT.ProceduralMaterial: ProceduralMaterial,
    CIT.ProceduralTexture: ProceduralTexture,
    CIT.Texture2DArray: Texture2DArray,
    CIT.CubemapArray: CubemapArray,
    CIT.OffMeshLink: OffMeshLink,
    CIT.OcclusionArea: OcclusionArea,
    CIT.Tree: Tree,
    CIT.NavMeshObsolete: NavMeshObsolete,
    CIT.NavMeshAgent: NavMeshAgent,
    CIT.NavMeshSettings: NavMeshSettings,
    CIT.LightProbesLegacy: LightProbes,
    CIT.ParticleSystem: ParticleSystem,
    CIT.ParticleSystemRenderer: ParticleSystemRenderer,
    CIT.ShaderVariantCollection: ShaderVariantCollection,
    CIT.LODGroup: LODGroup,
    CIT.BlendTree: BlendTree,
    CIT.Motion: Motion,
    CIT.NavMeshObstacle: NavMeshObstacle,
    CIT.SortingGroup: SortingGroup,
    CIT.SpriteRenderer: SpriteRenderer,
    CIT.Sprite: Sprite,
    CIT.CachedSpriteAtlas: CachedSpriteAtlas,
    CIT.ReflectionProbe: ReflectionProbe,
    CIT.ReflectionProbes: ReflectionProbe,
    CIT.Terrain: Terrain,
    CIT.LightProbeGroup: LightProbeGroup,
    CIT.AnimatorOverrideController: AnimatorOverrideController,
    CIT.CanvasRenderer: CanvasRenderer,
    CIT.Canvas: Canvas,
    CIT.RectTransform: RectTransform,
    CIT.CanvasGroup: CanvasGroup,
    CIT.BillboardAsset: BillboardAsset,
    CIT.BillboardRenderer: BillboardRenderer,
    CIT.SpeedTreeWindAsset: SpeedTreeWindAsset,
    CIT.AnchoredJoint2D: AnchoredJoint2D,
    CIT.Joint2D: Joint2D,
    CIT.SpringJoint2D: SpringJoint2D,
    CIT.DistanceJoint2D: DistanceJoint2D,
    CIT.HingeJoint2D: HingeJoint2D,
    CIT.SliderJoint2D: SliderJoint2D,
    CIT.WheelJoint2D: WheelJoint2D,
    CIT.ClusterInputManager: ClusterInputManager,
    CIT.BaseVideoTexture: BaseVideoTexture,
    CIT.NavMeshData: NavMeshData,
    CIT.AudioMixer: AudioMixer,
    CIT.AudioMixerController: AudioMixerController,
    CIT.AudioMixerGroupController: AudioMixerGroupController,
    CIT.AudioMixerEffectController: AudioMixerEffectController,
    CIT.AudioMixerSnapshotController: AudioMixerSnapshotController,
    CIT.PhysicsUpdateBehaviour2D: PhysicsUpdateBehaviour2D,
    CIT.ConstantForce2D: ConstantForce2D,
    CIT.Effector2D: Effector2D,
    CIT.AreaEffector2D: AreaEffector2D,
    CIT.PointEffector2D: PointEffector2D,
    CIT.PlatformEffector2D: PlatformEffector2D,
    CIT.SurfaceEffector2D: SurfaceEffector2D,
    CIT.BuoyancyEffector2D: BuoyancyEffector2D,
    CIT.RelativeJoint2D: RelativeJoint2D,
    CIT.FixedJoint2D: FixedJoint2D,
    CIT.FrictionJoint2D: FrictionJoint2D,
    CIT.TargetJoint2D: TargetJoint2D,
    CIT.LightProbes: LightProbes,
    CIT.LightProbeProxyVolume: LightProbeProxyVolume,
    CIT.SampleClip: SampleClip,
    CIT.AudioMixerSnapshot: AudioMixerSnapshot,
    CIT.AudioMixerGroup: AudioMixerGroup,
    CIT.NScreenBridge: NScreenBridge,
    CIT.AssetBundleManifest: AssetBundleManifest,
    CIT.UnityAdsManager: UnityAdsManager,
    CIT.RuntimeInitializeOnLoadManager: RuntimeInitializeOnLoadManager,
    CIT.CloudWebServicesManager: CloudWebServicesManager,
    CIT.UnityAnalyticsManager: UnityAnalyticsManager,
    CIT.CrashReportManager: CrashReportManager,
    CIT.PerformanceReportingManager: PerformanceReportingManager,
    CIT.UnityConnectSettings: UnityConnectSettings,
    CIT.AvatarMask: AvatarMask,
    CIT.PlayableDirector: PlayableDirector,
    CIT.VideoPlayer: VideoPlayer,
    CIT.VideoClip: VideoClip,
    CIT.ParticleSystemForceField: ParticleSystemForceField,
    CIT.SpriteMask: SpriteMask,
    CIT.WorldAnchor: WorldAnchor,
    CIT.OcclusionCullingData: OcclusionCullingData,
    # CIT.SmallestEditorClassID: SmallestEditorClassID,
    CIT.PrefabInstance: PrefabInstance,
    CIT.EditorExtensionImpl: EditorExtensionImpl,
    CIT.AssetImporter: AssetImporter,
    CIT.AssetDatabaseV1: AssetDatabaseV1,
    CIT.Mesh3DSImporter: Mesh3DSImporter,
    CIT.TextureImporter: TextureImporter,
    CIT.ShaderImporter: ShaderImporter,
    CIT.ComputeShaderImporter: ComputeShaderImporter,
    CIT.AudioImporter: AudioImporter,
    CIT.HierarchyState: HierarchyState,
    CIT.GUIDSerializer: GUIDSerializer,
    CIT.AssetMetaData: AssetMetaData,
    CIT.DefaultAsset: DefaultAsset,
    CIT.DefaultImporter: DefaultImporter,
    CIT.TextScriptImporter: TextScriptImporter,
    CIT.SceneAsset: SceneAsset,
    CIT.NativeFormatImporter: NativeFormatImporter,
    CIT.MonoImporter: MonoImporter,
    CIT.AssetServerCache: AssetServerCache,
    CIT.LibraryAssetImporter: LibraryAssetImporter,
    CIT.ModelImporter: ModelImporter,
    CIT.FBXImporter: FBXImporter,
    CIT.TrueTypeFontImporter: TrueTypeFontImporter,
    CIT.MovieImporter: MovieImporter,
    CIT.EditorBuildSettings: EditorBuildSettings,
    CIT.DDSImporter: DDSImporter,
    CIT.InspectorExpandedState: InspectorExpandedState,
    CIT.AnnotationManager: AnnotationManager,
    CIT.PluginImporter: PluginImporter,
    CIT.EditorUserBuildSettings: EditorUserBuildSettings,
    CIT.PVRImporter: PVRImporter,
    CIT.ASTCImporter: ASTCImporter,
    CIT.KTXImporter: KTXImporter,
    CIT.IHVImageFormatImporter: IHVImageFormatImporter,
    CIT.AnimatorStateTransition: AnimatorStateTransition,
    CIT.AnimatorState: AnimatorState,
    CIT.HumanTemplate: HumanTemplate,
    CIT.AnimatorStateMachine: AnimatorStateMachine,
    CIT.PreviewAnimationClip: PreviewAnimationClip,
    CIT.AnimatorTransition: AnimatorTransition,
    CIT.SpeedTreeImporter: SpeedTreeImporter,
    CIT.AnimatorTransitionBase: AnimatorTransitionBase,
    CIT.SubstanceImporter: SubstanceImporter,
    CIT.LightmapParameters: LightmapParameters,
    CIT.LightingDataAsset: LightingDataAsset,
    # CIT.GISRaster: GISRaster,
    # CIT.GISRasterImporter: GISRasterImporter,
    # CIT.CadImporter: CadImporter,
    CIT.SketchUpImporter: SketchUpImporter,
    CIT.BuildReport: BuildReport,
    CIT.PackedAssets: PackedAssets,
    CIT.VideoClipImporter: VideoClipImporter,
    # CIT.ActivationLogComponent: ActivationLogComponent,
    # CIT.int: int,
    # CIT.bool: bool,
    # CIT.float: float,
    CIT.MonoObject: MonoObject,
    CIT.Collision: Collision,
    CIT.Vector3f: Vector3f,
    CIT.RootMotionData: RootMotionData,
    CIT.Collision2D: Collision2D,
    CIT.AudioMixerLiveUpdateFloat: AudioMixerLiveUpdateFloat,
    CIT.AudioMixerLiveUpdateBool: AudioMixerLiveUpdateBool,
    CIT.Polygon2D: Polygon2D,
    # CIT.void: void,
    CIT.TilemapCollider2D: TilemapCollider2D,
    CIT.AssetImporterLog: AssetImporterLog,
    CIT.VFXRenderer: VFXRenderer,
    CIT.SerializableManagedRefTestClass: SerializableManagedRefTestClass,
    CIT.Grid: Grid,
    CIT.ScenesUsingAssets: ScenesUsingAssets,
    CIT.ArticulationBody: ArticulationBody,
    CIT.Preset: Preset,
    CIT.EmptyObject: EmptyObject,
    CIT.IConstraint: IConstraint,
    CIT.TestObjectWithSpecialLayoutOne: TestObjectWithSpecialLayoutOne,
    CIT.AssemblyDefinitionReferenceImporter: AssemblyDefinitionReferenceImporter,
    CIT.SiblingDerived: SiblingDerived,
    CIT.TestObjectWithSerializedMapStringNonAlignedStruct: TestObjectWithSerializedMapStringNonAlignedStruct,
    CIT.SubDerived: SubDerived,
    CIT.AssetImportInProgressProxy: AssetImportInProgressProxy,
    CIT.PluginBuildInfo: PluginBuildInfo,
    CIT.EditorProjectAccess: EditorProjectAccess,
    CIT.PrefabImporter: PrefabImporter,
    CIT.TestObjectWithSerializedArray: TestObjectWithSerializedArray,
    CIT.TestObjectWithSerializedAnimationCurve: TestObjectWithSerializedAnimationCurve,
    CIT.TilemapRenderer: TilemapRenderer,
    CIT.ScriptableCamera: ScriptableCamera,
    CIT.SpriteAtlasAsset: SpriteAtlasAsset,
    CIT.SpriteAtlasDatabase: SpriteAtlasDatabase,
    CIT.AudioBuildInfo: AudioBuildInfo,
    CIT.CachedSpriteAtlasRuntimeData: CachedSpriteAtlasRuntimeData,
    CIT.RendererFake: RendererFake,
    CIT.AssemblyDefinitionReferenceAsset: AssemblyDefinitionReferenceAsset,
    CIT.BuiltAssetBundleInfoSet: BuiltAssetBundleInfoSet,
    CIT.SpriteAtlas: SpriteAtlas,
    CIT.RayTracingShaderImporter: RayTracingShaderImporter,
    CIT.RayTracingShader: RayTracingShader,
    CIT.LightingSettings: LightingSettings,
    CIT.PlatformModuleSetup: PlatformModuleSetup,
    CIT.VersionControlSettings: VersionControlSettings,
    CIT.AimConstraint: AimConstraint,
    CIT.VFXManager: VFXManager,
    CIT.VisualEffectSubgraph: VisualEffectSubgraph,
    CIT.VisualEffectSubgraphOperator: VisualEffectSubgraphOperator,
    CIT.VisualEffectSubgraphBlock: VisualEffectSubgraphBlock,
    CIT.LocalizationImporter: LocalizationImporter,
    CIT.Derived: Derived,
    CIT.PropertyModificationsTargetTestObject: PropertyModificationsTargetTestObject,
    CIT.ReferencesArtifactGenerator: ReferencesArtifactGenerator,
    CIT.AssemblyDefinitionAsset: AssemblyDefinitionAsset,
    CIT.SceneVisibilityState: SceneVisibilityState,
    CIT.LookAtConstraint: LookAtConstraint,
    CIT.SpriteAtlasImporter: SpriteAtlasImporter,
    CIT.MultiArtifactTestImporter: MultiArtifactTestImporter,
    CIT.GameObjectRecorder: GameObjectRecorder,
    CIT.LightingDataAssetParent: LightingDataAssetParent,
    CIT.PresetManager: PresetManager,
    CIT.TestObjectWithSpecialLayoutTwo: TestObjectWithSpecialLayoutTwo,
    CIT.StreamingManager: StreamingManager,
    CIT.LowerResBlitTexture: LowerResBlitTexture,
    CIT.StreamingController: StreamingController,
    CIT.RenderPassAttachment: RenderPassAttachment,
    CIT.TestObjectVectorPairStringBool: TestObjectVectorPairStringBool,
    CIT.GridLayout: GridLayout,
    CIT.AssemblyDefinitionImporter: AssemblyDefinitionImporter,
    CIT.ParentConstraint: ParentConstraint,
    CIT.FakeComponent: FakeComponent,
    CIT.PositionConstraint: PositionConstraint,
    CIT.RotationConstraint: RotationConstraint,
    CIT.ScaleConstraint: ScaleConstraint,
    CIT.Tilemap: Tilemap,
    CIT.PackageManifest: PackageManifest,
    CIT.PackageManifestImporter: PackageManifestImporter,
    CIT.TerrainLayer: TerrainLayer,
    CIT.SpriteShapeRenderer: SpriteShapeRenderer,
    CIT.NativeObjectType: NativeObjectType,
    CIT.TestObjectWithSerializedMapStringBool: TestObjectWithSerializedMapStringBool,
    CIT.SerializableManagedHost: SerializableManagedHost,
    CIT.VisualEffectAsset: VisualEffectAsset,
    CIT.VisualEffectImporter: VisualEffectImporter,
    CIT.VisualEffectResource: VisualEffectResource,
    CIT.VisualEffectObject: VisualEffectObject,
    CIT.VisualEffect: VisualEffect,
    CIT.LocalizationAsset: LocalizationAsset,
    CIT.ScriptedImporter: ScriptedImporter,
}

__all__ = ["ClassIDTypeToClassMap"]



from enum import IntEnum


class ShaderCompilerPlatform(IntEnum):
    kShaderCompPlatformNone = -1
    kShaderCompPlatformGL = 0
    kShaderCompPlatformD3D9 = 1
    kShaderCompPlatformXbox360 = 2
    kShaderCompPlatformPS3 = 3
    kShaderCompPlatformD3D11 = 4
    kShaderCompPlatformGLES20 = 5
    kShaderCompPlatformNaCl = 6
    kShaderCompPlatformFlash = 7
    kShaderCompPlatformD3D11_9x = 8
    kShaderCompPlatformGLES3Plus = 9
    kShaderCompPlatformPSP2 = 10
    kShaderCompPlatformPS4 = 11
    kShaderCompPlatformXboxOne = 12
    kShaderCompPlatformPSM = 13
    kShaderCompPlatformMetal = 14
    kShaderCompPlatformOpenGLCore = 15
    kShaderCompPlatformN3DS = 16
    kShaderCompPlatformWiiU = 17
    kShaderCompPlatformVulkan = 18
    kShaderCompPlatformSwitch = 19
    kShaderCompPlatformXboxOneD3D12 = 20



# TODO: implement encryption for saving files
import re
from collections import namedtuple
from typing import Optional, Union, cast

from .. import config
from ..enums import ArchiveFlags, ArchiveFlagsOld, CompressionFlags
from ..helpers import ArchiveStorageManager, CompressionHelper
from ..helpers.UnityVersion import UnityVersion
from ..streams import EndianBinaryReader, EndianBinaryWriter
from . import File

BlockInfo = namedtuple("BlockInfo", "uncompressedSize compressedSize flags")
DirectoryInfoFS = namedtuple("DirectoryInfoFS", "offset size flags path")
reVersion = re.compile(r"(\d+)\.(\d+)\.(\d+)\w.+")


class BundleFile(File.File):
    format: int
    is_changed: bool
    signature: str
    version_engine: str
    version_player: str
    dataflags: Union[ArchiveFlags, ArchiveFlagsOld]
    decryptor: Optional[ArchiveStorageManager.ArchiveStorageDecryptor] = None
    _uses_block_alignment: bool = False

    def __init__(
        self,
        reader: EndianBinaryReader,
        parent: File,
        name: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(parent=parent, name=name, **kwargs)
        signature = self.signature = reader.read_string_to_null()
        self.version = reader.read_u_int()
        self.version_player = reader.read_string_to_null()
        self.version_engine = reader.read_string_to_null()

        if signature == "UnityArchive":
            raise NotImplementedError("BundleFile - UnityArchive")
        elif signature in ["UnityWeb", "UnityRaw"]:
            m_DirectoryInfo, blocksReader = self.read_web_raw(reader)
        elif signature == "UnityFS":
            m_DirectoryInfo, blocksReader = self.read_fs(reader)
        else:
            raise NotImplementedError(f"Unknown Bundle signature: {signature}")

        self.read_files(blocksReader, m_DirectoryInfo)

    def read_web_raw(self, reader: EndianBinaryReader):
        # def read_header_and_blocks_info(self, reader:EndianBinaryReader):
        version = self.version
        if version >= 4:
            self._hash = reader.read_bytes(16)
            self.crc = reader.read_u_int()

        minimumStreamedBytes = reader.read_u_int()  # noqa: F841
        headerSize = reader.read_u_int()
        numberOfLevelsToDownloadBeforeStreaming = reader.read_u_int()  # noqa: F841
        levelCount = reader.read_int()
        reader.Position += 4 * 2 * (levelCount - 1)

        compressedSize = reader.read_u_int()
        uncompressedSize = reader.read_u_int()  # noqa: F841

        if version >= 2:
            completeFileSize = reader.read_u_int()  # noqa: F841

        if version >= 3:
            fileInfoHeaderSize = reader.read_u_int()  # noqa: F841

        reader.Position = headerSize

        uncompressedBytes = reader.read_bytes(compressedSize)
        if self.signature == "UnityWeb":
            uncompressedBytes = CompressionHelper.decompress_lzma(uncompressedBytes, True)

        blocksReader = EndianBinaryReader(uncompressedBytes, offset=headerSize)
        nodesCount = blocksReader.read_int()
        m_DirectoryInfo = [
            File.DirectoryInfo(
                blocksReader.read_string_to_null(),  # path
                blocksReader.read_u_int(),  # offset
                blocksReader.read_u_int(),  # size
            )
            for _ in range(nodesCount)
        ]

        return m_DirectoryInfo, blocksReader

    def read_fs(self, reader: EndianBinaryReader):
        size = reader.read_long()  # noqa: F841

        # header
        compressedSize = reader.read_u_int()
        uncompressedSize = reader.read_u_int()
        dataflagsValue = reader.read_u_int()

        version = self.parse_version()
        # https://issuetracker.unity3d.com/issues/files-within-assetbundles-do-not-start-on-aligned-boundaries-breaking-patching-on-nintendo-switch
        # Unity CN introduced encryption before the alignment fix was introduced.
        # Unity CN used the same flag for the encryption as later on the alignment fix,
        # so we have to check the version to determine the correct flag set.
        if (
            version < (2020,)
            or (version[0] == 2020 and version < (2020, 3, 34))
            or (version[0] == 2021 and version < (2021, 3, 2))
            or (version[0] == 2022 and version < (2022, 1, 1))
        ):
            self.dataflags = ArchiveFlagsOld(dataflagsValue)
        else:
            self.dataflags = ArchiveFlags(dataflagsValue)

        if self.dataflags & self.dataflags.UsesAssetBundleEncryption:
            self.decryptor = ArchiveStorageManager.ArchiveStorageDecryptor(reader)

        # check if we need to align the reader
        # - align to 16 bytes and check if all are 0
        # - if not, reset the reader to the previous position
        if self.version >= 7:
            reader.align_stream(16)
            self._uses_block_alignment = True
        elif version >= (2019, 4):
            pre_align = reader.Position
            align_data = reader.read((16 - pre_align % 16) % 16)
            if any(align_data):
                reader.Position = pre_align
            else:
                self._uses_block_alignment = True

        start = reader.Position
        if self.dataflags & ArchiveFlags.BlocksInfoAtTheEnd:  # kArchiveBlocksInfoAtTheEnd
            reader.Position = reader.Length - compressedSize
            blocksInfoBytes = reader.read_bytes(compressedSize)
            reader.Position = start
        else:  # 0x40 kArchiveBlocksAndDirectoryInfoCombined
            blocksInfoBytes = reader.read_bytes(compressedSize)

        blocksInfoBytes = self.decompress_data(blocksInfoBytes, uncompressedSize, self.dataflags)
        blocksInfoReader = EndianBinaryReader(blocksInfoBytes, offset=start)

        uncompressedDataHash = blocksInfoReader.read_bytes(16)  # noqa: F841
        blocksInfoCount = blocksInfoReader.read_int()

        m_BlocksInfo = [
            BlockInfo(
                blocksInfoReader.read_u_int(),  # uncompressedSize
                blocksInfoReader.read_u_int(),  # compressedSize
                blocksInfoReader.read_u_short(),  # flags
            )
            for _ in range(blocksInfoCount)
        ]

        nodesCount = blocksInfoReader.read_int()
        m_DirectoryInfo = [
            DirectoryInfoFS(
                blocksInfoReader.read_long(),  # offset
                blocksInfoReader.read_long(),  # size
                blocksInfoReader.read_u_int(),  # flags
                blocksInfoReader.read_string_to_null(),  # path
            )
            for _ in range(nodesCount)
        ]

        if m_BlocksInfo:
            self._block_info_flags = m_BlocksInfo[0].flags

        if isinstance(self.dataflags, ArchiveFlags) and self.dataflags & ArchiveFlags.BlockInfoNeedPaddingAtStart:
            reader.align_stream(16)

        blocksReader = EndianBinaryReader(
            b"".join(
                self.decompress_data(
                    reader.read_bytes(blockInfo.compressedSize),
                    blockInfo.uncompressedSize,
                    blockInfo.flags,
                    i,
                )
                for i, blockInfo in enumerate(m_BlocksInfo)
            ),
            offset=(blocksInfoReader.real_offset()),
        )

        return m_DirectoryInfo, blocksReader

    def save(self, packer=None):
        """
        Rewrites the BundleFile and returns it as bytes object.

        packer:
            can be either one of the following strings
            or tuple consisting of (block_info_flag, data_flag)
            allowed strings:
                none - no compression, default, safest bet
                lz4 - lz4 compression
                original - uses the original flags
        """
        # file_header
        #     signature         (string_to_null)
        #     format            (int)
        #     version_player    (string_to_null)
        #     version_engine    (string_to_null)
        writer = EndianBinaryWriter()

        writer.write_string_to_null(self.signature)
        writer.write_u_int(self.version)
        writer.write_string_to_null(self.version_player)
        writer.write_string_to_null(self.version_engine)

        if self.signature == "UnityArchive":
            raise NotImplementedError("BundleFile - UnityArchive")
        elif self.signature in ["UnityWeb", "UnityRaw"]:
            self.save_web_raw(writer)
        elif self.signature == "UnityFS":
            if not packer or packer == "none":
                self.save_fs(writer, 64, 64)
            elif packer == "original":
                self.save_fs(
                    writer,
                    data_flag=self.dataflags,
                    block_info_flag=self._block_info_flags,
                )
            elif packer == "lz4":
                self.save_fs(writer, data_flag=194, block_info_flag=2)
            elif packer == "lzma":
                self.save_fs(writer, data_flag=65, block_info_flag=1)
            elif isinstance(packer, tuple):
                self.save_fs(writer, *packer)
            else:
                raise NotImplementedError("UnityFS - Packer:", packer)
        return writer.bytes

    def save_fs(self, writer: EndianBinaryWriter, data_flag: int, block_info_flag: int):
        # header
        # compressed blockinfo (block details & directionary)
        # compressed assets

        # 0b1000000 / 0b11000000 | 64 / 192 - uncompressed
        # 0b11000010 | 194 - lz4
        # block_info_flag

        # 0 / 0b1000000 | 0 / 64 - uncompressed
        # 0b1   | 1 - lzma
        # 0b10  | 2 - lz4
        # 0b11  | 3 - lz4hc [not implemented]
        # 0b100 | 4 - lzham [not implemented]
        # data_flag

        # header:
        #     bundle_size       (long)
        #     compressed_size   (int)
        #     uncompressed_size (int)
        #     flag              (int)
        #     ?padding?         (bool)
        #   This will be written at the end,
        #   because the size can only be calculated after the data compression,

        # block_info:
        #     *flag & 0x80 ? at the end : right after header
        #     *decompression via flag & 0x3F
        #     *read compressed_size -> uncompressed_size
        #     0x10 offset
        #     *read blocks infos of the data stream
        #     count                 (int)
        #     (
        #         uncompressed_size (uint)
        #         compressed_size   (uint)
        #         flag              (short)
        #     )
        #     *decompression via info.flag & 0x3F

        #     *afterwards the file positions
        #     file_count    (int)
        #     (
        #         offset    (long)
        #         size      (long)
        #         flag      (int)
        #         name      (string_to_null)
        #     )

        # file list & file data
        # prep nodes and build up block data
        data_writer = EndianBinaryWriter()
        files = [
            (
                name,
                f.flags,
                data_writer.write_bytes(
                    f.bytes if isinstance(f, (EndianBinaryReader, EndianBinaryWriter)) else f.save()
                ),
            )
            for name, f in self.files.items()
        ]

        file_data = data_writer.bytes
        data_writer.dispose()

        # remove encryption flag, as encryption isn't done
        if block_info_flag & self.dataflags.UsesAssetBundleEncryption:
            block_info_flag ^= self.dataflags.UsesAssetBundleEncryption
        if data_flag & self.dataflags.UsesAssetBundleEncryption:
            data_flag ^= self.dataflags.UsesAssetBundleEncryption

        file_data, block_info = CompressionHelper.chunk_based_compress(file_data, block_info_flag)

        # write the block_info
        # uncompressedDataHash
        block_writer = EndianBinaryWriter(b"\x00" * 0x10)
        # data block info
        block_writer.write_int(len(block_info))
        for block_uncompressed_size, block_compressed_size, block_flag in block_info:
            # uncompressed size
            block_writer.write_u_int(block_uncompressed_size)
            # compressed size
            block_writer.write_u_int(block_compressed_size)
            # flag
            block_writer.write_u_short(block_flag)

        # file block info
        if not data_flag & 0x40:
            raise NotImplementedError("UnityPy always writes DirectoryInfo, so data_flag must include 0x40")
        # file count
        block_writer.write_int(len(files))
        offset = 0
        for f_name, f_flag, f_len in files:
            # offset
            block_writer.write_long(offset)
            # size
            block_writer.write_long(f_len)
            offset += f_len
            # flag
            block_writer.write_u_int(f_flag)
            # name
            block_writer.write_string_to_null(f_name)

        # compress the block data
        block_data = block_writer.bytes
        block_writer.dispose()

        uncompressed_block_data_size = len(block_data)

        switch = data_flag & 0x3F
        if switch in CompressionHelper.COMPRESSION_MAP:
            block_data = CompressionHelper.COMPRESSION_MAP[switch](block_data)
        else:
            raise NotImplementedError(f"No compression function in the CompressionHelper.COMPRESSION_MAP for {switch}")

        compressed_block_data_size = len(block_data)

        # write the header info
        ## file size - 0 for now, will be set at the end
        writer_header_pos = writer.Position
        writer.write_long(0)
        # compressed blockInfoBytes size
        writer.write_u_int(compressed_block_data_size)
        # uncompressed size
        writer.write_u_int(uncompressed_block_data_size)
        # compression and file layout flag
        writer.write_u_int(data_flag)

        if self._uses_block_alignment:
            # UnityFS\x00 - 8
            # size 8
            # comp sizes 4+4
            # flag 4
            # sum : 28 -> +8 alignment
            writer.align_stream(16)

        if data_flag & 0x80:  # at end of file
            if data_flag & 0x200:
                writer.align_stream(16)
            writer.write(file_data)
            writer.write(block_data)
        else:
            writer.write(block_data)
            if data_flag & 0x200:
                writer.align_stream(16)
            writer.write(file_data)

        writer_end_pos = writer.Position
        writer.Position = writer_header_pos
        # correct file size
        writer.write_long(writer_end_pos)
        writer.Position = writer_end_pos

    def save_web_raw(self, writer: EndianBinaryWriter):
        # (version >= 4) hash
        # (version >= 4) crc
        # minimumStreamedBytes
        # headerSize
        # numberOfLevelsToDownloadBeforeStreaming
        # levelCount
        # compressedSize * levelCount
        # uncompressedSize * levelCount
        # (version >= 2) completeFileSize
        # (version >= 3) file_info_header_size
        # compressed assets

        if self.version > 3:
            raise NotImplementedError("Saving Unity Web bundles with version > 3 is not supported")

        # Calculate fileInfoHeaderSize for set offsets
        file_info_header_size = 4  # for nodesCount

        for file_name in self.files.keys():
            file_info_header_size += len(file_name.encode()) + 1  # +1 for null terminator
            file_info_header_size += 4 * 2  # 4 bytes each for offset and size

        file_info_header_padding_size = 4 - (file_info_header_size % 4) if file_info_header_size % 4 != 0 else 0
        file_info_header_size += file_info_header_padding_size

        # Prepare directory info
        directory_info_writer = EndianBinaryWriter()
        directory_info_writer.write_int(len(self.files))  # nodesCount

        file_content_writer = EndianBinaryWriter()
        current_offset = file_info_header_size

        for file_name, f in self.files.items():
            directory_info_writer.write_string_to_null(file_name)
            directory_info_writer.write_u_int(current_offset)

            # Get file content
            if isinstance(f, (EndianBinaryReader, EndianBinaryWriter)):
                file_data = f.bytes
            else:
                file_data = f.save()

            file_size = len(file_data)
            directory_info_writer.write_u_int(file_size)

            file_content_writer.write_bytes(file_data)
            current_offset += file_size

        directory_info_writer.write(b"\x00" * file_info_header_padding_size)
        uncompressed_directory_info = directory_info_writer.bytes
        uncompressed_file_content = file_content_writer.bytes

        # Combine directory info and file content
        uncompressed_content = uncompressed_directory_info + uncompressed_file_content
        compressed_content = uncompressed_content
        if self.signature == "UnityWeb":
            compressed_content = CompressionHelper.compress_lzma(uncompressed_content, True)

        # Write header
        header_size = writer.Position + 24  # assuming levelCount = 1
        if self.version >= 2:
            header_size += 4
        if self.version >= 3:
            header_size += 4
        if self.version >= 4:
            header_size += 20
        # pad to multiple of 4
        header_size = (header_size + 3) & ~3

        if self.version >= 4:
            writer.write_bytes(self._hash)
            writer.write_u_int(self.crc)

        writer.write_u_int(header_size + len(compressed_content))  # minimumStreamedBytes (same as completeFileSize)
        writer.write_u_int(header_size)  # headerSize
        writer.write_u_int(1)  # numberOfLevelsToDownloadBeforeStreaming (always 1)
        writer.write_int(1)  # levelCount (always 1)

        writer.write_u_int(len(compressed_content))  # compressedSize
        writer.write_u_int(len(uncompressed_content))  # uncompressedSize

        if self.version >= 2:
            writer.write_u_int(header_size + len(compressed_content))  # completeFileSize

        if self.version >= 3:
            writer.write_u_int(file_info_header_size)  # file_info_header_size

        # align header
        writer.align_stream(4)

        # Write compressed content
        writer.write(compressed_content)

    def decompress_data(
        self,
        compressed_data: bytes,
        uncompressed_size: int,
        flags: Union[int, ArchiveFlags, ArchiveFlagsOld],
        index: int = 0,
    ) -> bytes:
        """
        Parameters
        ----------
        compressed_data : bytes
            The compressed data.
        uncompressed_size : int
            The uncompressed size of the data.
        flags : int
            The flags of the data.

        Returns
        -------
        bytes
            The decompressed data."""
        comp_flag = CompressionFlags(flags & ArchiveFlags.CompressionTypeMask)

        if self.decryptor is not None and flags & 0x100 and comp_flag != CompressionFlags.NONE:
            compressed_data = self.decryptor.decrypt_block(compressed_data, index)

        if comp_flag in CompressionHelper.DECOMPRESSION_MAP:
            return cast(
                bytes,
                CompressionHelper.DECOMPRESSION_MAP[comp_flag](compressed_data, uncompressed_size),
            )
        else:
            raise ValueError(f"Unknown compression! flag: {flags}, compression flag: {comp_flag.value}")

    def parse_version(self) -> UnityVersion:
        """Returns the version as a tuple."""
        version = None
        version_str = self.version_engine
        try:
            version = UnityVersion.from_str(version_str)
        except ValueError:
            pass

        if version is None or version.major == 0:
            version_str = config.get_fallback_version()
            version = UnityVersion.from_str(version_str)

        return version



import gzip
import lzma
import struct
from typing import Callable, Dict, Tuple, Union

import brotli
import lz4.block

from ..enums.BundleFile import CompressionFlags

ByteString = Union[bytes, bytearray, memoryview]
GZIP_MAGIC: bytes = b"\x1f\x8b"
BROTLI_MAGIC: bytes = b"brotli"


# LZMA
def decompress_lzma(data: ByteString, read_decompressed_size: bool = False) -> bytes:
    """decompresses lzma-compressed data

    :param data: compressed data
    :type data: ByteString
    :raises _lzma.LZMAError: Compressed data ended before the end-of-stream marker was reached
    :return: uncompressed data
    :rtype: bytes
    """
    props, dict_size = struct.unpack("<BI", data[:5])
    lc = props % 9
    remainder = props // 9
    pb = remainder // 5
    lp = remainder % 5
    dec = lzma.LZMADecompressor(
        format=lzma.FORMAT_RAW,
        filters=[
            {
                "id": lzma.FILTER_LZMA1,
                "dict_size": dict_size,
                "lc": lc,
                "lp": lp,
                "pb": pb,
            }
        ],
    )
    data_offset = 13 if read_decompressed_size else 5
    return dec.decompress(data[data_offset:])


def compress_lzma(data: ByteString, write_decompressed_size: bool = False) -> bytes:
    """compresses data via lzma (unity specific)
    The current static settings may not be the best solution,
    but they are the most commonly used values and should therefore be enough for the time being.

    :param data: uncompressed data
    :type data: ByteString
    :return: compressed data
    :rtype: bytes
    """
    dict_size = 0x800000  # 1 << 23
    compressor = lzma.LZMACompressor(
        format=lzma.FORMAT_RAW,
        filters=[
            {
                "id": lzma.FILTER_LZMA1,
                "dict_size": dict_size,
                "lc": 3,
                "lp": 0,
                "pb": 2,
                "mode": lzma.MODE_NORMAL,
                "mf": lzma.MF_BT4,
                "nice_len": 123,
            }
        ],
    )

    compressed_data = compressor.compress(data) + compressor.flush()
    cdl = len(compressed_data)
    if write_decompressed_size:
        return struct.pack(f"<BIQ{cdl}s", 0x5D, dict_size, len(data), compressed_data)
    else:
        return struct.pack(f"<BI{cdl}s", 0x5D, dict_size, compressed_data)


# LZ4
def decompress_lz4(data: ByteString, uncompressed_size: int) -> bytes:  # LZ4M/LZ4HC
    """decompresses lz4-compressed data

    :param data: compressed data
    :type data: ByteString
    :param uncompressed_size: size of the uncompressed data
    :type uncompressed_size: int
    :raises _block.LZ4BlockError: Decompression failed: corrupt input or insufficient space in destination buffer.
    :return: uncompressed data
    :rtype: bytes
    """
    return lz4.block.decompress(data, uncompressed_size)


def compress_lz4(data: ByteString) -> bytes:  # LZ4M/LZ4HC
    """compresses data via lz4.block

    :param data: uncompressed data
    :type data: ByteString
    :return: compressed data
    :rtype: bytes
    """
    return lz4.block.compress(data, mode="high_compression", compression=9, store_size=False)


# Brotli
def decompress_brotli(data: ByteString) -> bytes:
    """decompresses brotli-compressed data

    :param data: compressed data
    :type data: ByteString
    :raises brotli.error: BrotliDecompress failed
    :return: uncompressed data
    :rtype: bytes
    """
    return brotli.decompress(data)


def compress_brotli(data: ByteString) -> bytes:
    """compresses data via brotli

    :param data: uncompressed data
    :type data: ByteString
    :return: compressed data
    :rtype: bytes
    """
    return brotli.compress(data)


# GZIP
def decompress_gzip(data: ByteString) -> bytes:
    """decompresses gzip-compressed data

    :param data: compressed data
    :type data: ByteString
    :raises OSError: Not a gzipped file
    :return: uncompressed data
    :rtype: bytes
    """
    return gzip.decompress(data)


def compress_gzip(data: ByteString) -> bytes:
    """compresses data via gzip
    The current static settings may not be the best solution,
    but they are the most commonly used values and should therefore be enough for the time being.

    :param data: uncompressed data
    :type data: ByteString
    :return: compressed data
    :rtype: bytes
    """
    return gzip.compress(data)


def chunk_based_compress(data: ByteString, block_info_flag: int) -> Tuple[ByteString, list]:
    """compresses AssetBundle data based on the block_info_flag
    LZ4/LZ4HC will be chunk-based compression

    :param data: uncompressed data
    :type data: ByteString
    :param block_info_flag: block info flag
    :type block_info_flag: int
    :return: compressed data and block info
    :rtype: tuple
    """
    switch = block_info_flag & 0x3F
    chunk_size = None
    compress_func = None
    if switch == 0:  # NONE
        return data, [(len(data), len(data), block_info_flag)]

    if switch in COMPRESSION_MAP:
        compress_func = COMPRESSION_MAP[switch]
    else:
        raise NotImplementedError(f"No compression function in the CompressionHelper.COMPRESSION_MAP for {switch}")

    if switch in COMPRESSION_CHUNK_SIZE_MAP:
        chunk_size = COMPRESSION_CHUNK_SIZE_MAP[switch]
    else:
        raise NotImplementedError(f"No chunk size in the CompressionHelper.COMPRESSION_CHUNK_SIZE_MAP for {switch}")

    block_info = []
    uncompressed_data_size = len(data)
    compressed_file_data = bytearray()
    p = 0
    while uncompressed_data_size > chunk_size:
        compressed_data = compress_func(data[p : p + chunk_size])
        if len(compressed_data) > chunk_size:
            compressed_file_data.extend(data[p : p + chunk_size])
            block_info.append(
                (
                    chunk_size,
                    chunk_size,
                    block_info_flag ^ switch,
                )
            )
        else:
            compressed_file_data.extend(compressed_data)
            block_info.append(
                (
                    chunk_size,
                    len(compressed_data),
                    block_info_flag,
                )
            )
        p += chunk_size
        uncompressed_data_size -= chunk_size
    if uncompressed_data_size > 0:
        compressed_data = compress_func(data[p:])
        if len(compressed_data) > uncompressed_data_size:
            compressed_file_data.extend(data[p:])
            block_info.append(
                (
                    uncompressed_data_size,
                    uncompressed_data_size,
                    block_info_flag ^ switch,
                )
            )
        else:
            compressed_file_data.extend(compressed_data)
            block_info.append(
                (
                    uncompressed_data_size,
                    len(compressed_data),
                    block_info_flag,
                )
            )
    return bytes(compressed_file_data), block_info


def decompress_lzham(data: ByteString, uncompressed_size: int) -> bytes:
    raise NotImplementedError("Custom compression or unimplemented LZHAM (removed by Unity) encountered!")


DECOMPRESSION_MAP: Dict[Union[int, CompressionFlags], Callable[[ByteString, int], ByteString]] = {
    CompressionFlags.NONE: lambda cd, _ucs: cd,
    CompressionFlags.LZMA: lambda cd, _ucs: decompress_lzma(cd),
    CompressionFlags.LZ4: decompress_lz4,
    CompressionFlags.LZ4HC: decompress_lz4,
    CompressionFlags.LZHAM: decompress_lzham,
}

COMPRESSION_MAP: Dict[Union[int, CompressionFlags], Callable[[ByteString], ByteString]] = {
    CompressionFlags.NONE: lambda cd: cd,
    CompressionFlags.LZMA: compress_lzma,
    CompressionFlags.LZ4: compress_lz4,
    CompressionFlags.LZ4HC: compress_lz4,
}

COMPRESSION_CHUNK_SIZE_MAP: Dict[Union[int, CompressionFlags], int] = {
    CompressionFlags.NONE: 0xFFFFFFFF,
    CompressionFlags.LZMA: 0xFFFFFFFF,
    CompressionFlags.LZ4: 0x00020000,
    CompressionFlags.LZ4HC: 0x00020000,
}


__all__ = (
    "compress_brotli",
    "compress_gzip",
    "compress_lz4",
    "compress_lzma",
    "decompress_brotli",
    "decompress_gzip",
    "decompress_lz4",
    "decompress_lzma",
    "decompress_lzham",
    "chunk_based_compress",
    "COMPRESSION_MAP",
    "DECOMPRESSION_MAP",
    "COMPRESSION_CHUNK_SIZE_MAP",
)



from __future__ import annotations

import re
from itertools import groupby
from typing import TYPE_CHECKING, List, Optional, Tuple, TypeVar, Union

from ..enums import (
    PassType,
    SerializedPropertyType,
    ShaderCompilerPlatform,
    ShaderGpuProgramType,
    TextureDimension,
)
from ..helpers import CompressionHelper
from ..streams import EndianBinaryReader

if TYPE_CHECKING:
    from ..classes import (
        SerializedPass,
        SerializedProperties,
        SerializedProperty,
        SerializedShader,
        SerializedShaderState,
        SerializedSubProgram,
        SerializedSubShader,
        SerializedTagMap,
        Shader,
    )

HEADER = """
//////////////////////////////////////////
//
// NOTE: This is *not* a valid shader file
//
///////////////////////////////////////////
"""[1:]

T = TypeVar("T")


def export_shader(m_Shader: Shader) -> str:
    if m_Shader.m_SubProgramBlob:  # 5.3 - 5.4
        decompressedBytes = CompressionHelper.decompress_lz4(
            bytes(m_Shader.m_SubProgramBlob), m_Shader.decompressedSize
        )

        blobReader = EndianBinaryReader(decompressedBytes)
        program = ShaderProgram(blobReader, m_Shader.object_reader.version)

        return HEADER + program.Export(bytes(m_Shader.m_Script).decode("utf8"))

    if m_Shader.compressedBlob:  # 5.5 and up
        return HEADER + ConvertSerializedShader(m_Shader)

    return HEADER + bytes(m_Shader.m_Script).decode("utf8")


def ConvertSerializedShader(m_Shader: Shader) -> str:
    shaderPrograms = []

    platformNumber = len(m_Shader.platforms)
    compressed_blob = bytes(m_Shader.compressedBlob)

    def get_entry(array: Union[List[T], List[List[T]]], index: int) -> T:
        item = array[index]
        if isinstance(item, List):
            return item[0]
        return item

    for i in range(platformNumber):
        if i >= len(m_Shader.compressedLengths) or i >= len(m_Shader.decompressedLengths):
            # m_Shader.platforms shouldn't be longer than m_shader.[de]compressedLengths, but it is
            break

        compressedSize = get_entry(m_Shader.compressedLengths, i)
        decompressedSize = get_entry(m_Shader.decompressedLengths, i)
        offset = get_entry(m_Shader.offsets, i)

        compressedBytes = compressed_blob[offset : offset + compressedSize]
        decompressedBytes = CompressionHelper.decompress_lz4(compressedBytes, decompressedSize)

        shaderPrograms.append(
            ShaderProgram(
                EndianBinaryReader(decompressedBytes, endian="<"),
                m_Shader.object_reader.version,
            )
        )

    return ConvertSerializedShaderParsedForm(m_Shader.m_ParsedForm, m_Shader.platforms, shaderPrograms)


def ConvertSerializedShaderParsedForm(
    m_ParsedForm: SerializedShader,
    platforms: List[int],
    shaderPrograms: List[ShaderProgram],
) -> str:
    sb: List[str] = [
        'Shader "{0}" {{\n'.format(m_ParsedForm.m_Name),
        ConvertSerializedProperties(m_ParsedForm.m_PropInfo),
        *[
            ConvertSerializedSubShader(m_SubShader, platforms, shaderPrograms)
            for m_SubShader in m_ParsedForm.m_SubShaders
        ],
    ]

    if m_ParsedForm.m_FallbackName:
        sb.append('Fall back "{0}"\n'.format(m_ParsedForm.m_FallbackName))

    if m_ParsedForm.m_CustomEditorName:
        sb.append('CustomEditor "{0}"\n'.format(m_ParsedForm.m_CustomEditorName))

    sb.append("}")
    return "".join(sb)


def ConvertSerializedSubShader(
    m_SubShader: SerializedSubShader,
    platforms: List[int],
    shaderPrograms: List[ShaderProgram],
) -> str:
    sb = ["SubShader {\n"]

    if m_SubShader.m_LOD != 0:
        sb.append(" LOD {0}\n".format(m_SubShader.m_LOD))

    sb.append(ConvertSerializedTagMap(m_SubShader.m_Tags, 1))

    sb.extend(ConvertSerializedPass(m_Passe, platforms, shaderPrograms) for m_Passe in m_SubShader.m_Passes)
    sb.append("}\n")
    return "".join(sb)


def ConvertSerializedPass(m_Passe: SerializedPass, platforms: List[int], shaderPrograms: List[ShaderProgram]) -> str:
    sb = []
    if m_Passe.m_Type == PassType.kPassTypeNormal:
        sb.append(" Pass ")
    elif m_Passe.m_Type == PassType.kPassTypeUse:
        sb.append(" UsePass ")
    elif m_Passe.m_Type == PassType.kPassTypeGrab:
        sb.append(" GrabPass ")

    if m_Passe.m_Type == PassType.kPassTypeUse:
        sb.append('"{0}"\n'.format(m_Passe.m_UseName))
    else:
        sb.append("{\n")

        if m_Passe.m_Type == PassType.kPassTypeGrab:
            if m_Passe.m_TextureName:
                sb.append(' "{0}"\n'.format(m_Passe.m_TextureName))

        else:
            sb.append(ConvertSerializedShaderState(m_Passe.m_State))

            if len(m_Passe.progVertex.m_SubPrograms) > 0:
                sb.append('Program "vp" {\n')
                sb.append(ConvertSerializedSubPrograms(m_Passe.progVertex.m_SubPrograms, platforms, shaderPrograms))
                sb.append("}\n")

            if len(m_Passe.progFragment.m_SubPrograms) > 0:
                sb.append('Program "fp" {\n')
                sb.append(ConvertSerializedSubPrograms(m_Passe.progFragment.m_SubPrograms, platforms, shaderPrograms))
                sb.append("}\n")

            if len(m_Passe.progGeometry.m_SubPrograms) > 0:
                sb.append('Program "gp" {\n')
                sb.append(ConvertSerializedSubPrograms(m_Passe.progGeometry.m_SubPrograms, platforms, shaderPrograms))
                sb.append("}\n")

            if len(m_Passe.progHull.m_SubPrograms) > 0:
                sb.append('Program "hp" {\n')
                sb.append(ConvertSerializedSubPrograms(m_Passe.progHull.m_SubPrograms, platforms, shaderPrograms))
                sb.append("}\n")

            if len(m_Passe.progDomain.m_SubPrograms) > 0:
                sb.append('Program "dp" {\n')
                sb.append(ConvertSerializedSubPrograms(m_Passe.progDomain.m_SubPrograms, platforms, shaderPrograms))
                sb.append("}\n")

        sb.append("}\n")

    return "".join(sb)


def ConvertSerializedSubPrograms(
    m_SubPrograms: List[SerializedSubProgram],
    platforms: List[int],
    shaderPrograms: List[ShaderProgram],
) -> str:
    sb = []

    def indexFunc(x: SerializedSubProgram):
        return x.m_BlobIndex

    def typeFunc(x: SerializedSubProgram):
        return x.m_GpuProgramType

    groups = groupby(sorted(m_SubPrograms, key=indexFunc), indexFunc)

    for _, _group in groups:
        group = list(_group)
        programs = groupby(sorted(group, key=typeFunc), typeFunc)

        for programKey, _programList in programs:
            subPrograms = list(_programList)
            isTier = len(subPrograms) > 1
            for i in range(len(platforms)):
                if i >= len(shaderPrograms):
                    # platforms shouldn't be longer than shaderPrograms, but it is
                    break

                platform = platforms[i]

                if CheckGpuProgramUsable(platform, programKey):
                    for subProgram in subPrograms:
                        sb.append('SubProgram "{0} '.format(GetPlatformString(platform)))

                        if isTier:
                            sb.append("hw_tier{0:02} ".format(subProgram.m_ShaderHardwareTier))

                        sb.append('" {\n')
                        sb.append(shaderPrograms[i].m_SubPrograms[subProgram.m_BlobIndex].Export())

                        sb.append("\n}\n")

                    break

    return "".join(sb)


def ConvertSerializedShaderState(m_State: SerializedShaderState) -> str:
    sb = []
    if m_State.m_Name:
        sb.append(' Name "{0}"\n'.format(m_State.m_Name))

    if m_State.m_LOD != 0:
        sb.append("  LOD {0}\n".format(m_State.m_LOD))

    sb.append(ConvertSerializedTagMap(m_State.m_Tags, 2))

    # sb.append(ConvertSerializedShaderRTBlendState(m_State.rtBlend))

    if m_State.alphaToMask.val > 0:
        sb.append(" AlphaToMask On\n")

    if m_State.zClip and m_State.zClip.val != 1:
        sb.append(" ZClip Off\n")

    if m_State.zTest.val != 4:
        sb.append(" ZTest ")
        if m_State.zTest.val == 0:  # kFuncDisabled
            sb.append("Off")
        elif m_State.zTest.val == 1:  # kFuncNever
            sb.append("Never")
        elif m_State.zTest.val == 2:  # kFuncLess
            sb.append("Less")
        elif m_State.zTest.val == 3:  # kFuncEqual
            sb.append("Equal")
        elif m_State.zTest.val == 5:  # kFuncGreater
            sb.append("Greater")
        elif m_State.zTest.val == 6:  # kFuncNotEqual
            sb.append("NotEqual")
        elif m_State.zTest.val == 7:  # kFuncGEqual
            sb.append("GEqual")
        elif m_State.zTest.val == 8:  # kFuncAlways
            sb.append("Always")

        sb.append("\n")

    if m_State.zWrite.val != 1:  # ZWrite On
        sb.append(" ZWrite Off\n")

    if m_State.culling.val != 2:  # Cull Back
        sb.append(" Cull ")
        if m_State.culling.val == 0:  # kCullOff
            sb.append("Off")
        elif m_State.culling.val == 1:  # kCullFront
            sb.append("Front")

        sb.append("\n")

    if m_State.offsetFactor.val != 0 or m_State.offsetUnits.val != 0:
        sb.append(" Offset {0}, {1}\n".format(m_State.offsetFactor.val, m_State.offsetUnits.val))

    # TODO Stencil

    # TODO Fog

    if m_State.lighting:
        sb.append("  Lighting {0}\n".format(m_State.lighting and "On" or "Off"))

    sb.append("  GpuProgramID {0}\n".format(m_State.gpuProgramID))
    return "".join(sb)


def ConvertSerializedShaderRTBlendState(rbBlend):
    # TODO Blend
    sb = []
    return "".join(sb)


def ConvertSerializedTagMap(m_Tags: SerializedTagMap, intent: int) -> str:
    if m_Tags.tags:
        return "".join(
            [
                " " * intent,
                "Tags { ",
                *[f'"{key}" = "{value}" ' for key, value in m_Tags.tags],
                "}\n",
            ]
        )
    return ""


def ConvertSerializedProperties(m_PropInfo: SerializedProperties) -> str:
    return "\n".join(
        [
            "Properties {\n",
            *[ConvertSerializedProperty(m_Prop) for m_Prop in m_PropInfo.m_Props],
            "}\n",
        ]
    )


def ConvertSerializedProperty(m_Prop: SerializedProperty) -> str:
    sb = ["[{0}] ".format(m_Attribute) for m_Attribute in m_Prop.m_Attributes]
    sb.append('{0} ("{1}", '.format(m_Prop.m_Name, m_Prop.m_Description))

    if m_Prop.m_Type == SerializedPropertyType.kColor:
        sb.append("Color")
    elif m_Prop.m_Type == SerializedPropertyType.kVector:
        sb.append("Vector")
    elif m_Prop.m_Type == SerializedPropertyType.kFloat:
        sb.append("Float")
    elif m_Prop.m_Type == SerializedPropertyType.kRange:
        sb.append("Range({0:g}, {1:g})".format(m_Prop.m_DefValue_1_, m_Prop.m_DefValue_2_))
    elif m_Prop.m_Type == SerializedPropertyType.kTexture:
        if m_Prop.m_DefTexture.m_TexDim == TextureDimension.kTexDimAny:
            sb.append("any")
        elif m_Prop.m_DefTexture.m_TexDim == TextureDimension.kTexDim2D:
            sb.append("2D")
        elif m_Prop.m_DefTexture.m_TexDim == TextureDimension.kTexDim3D:
            sb.append("3D")
        elif m_Prop.m_DefTexture.m_TexDim == TextureDimension.kTexDimCUBE:
            sb.append("Cube")
        elif m_Prop.m_DefTexture.m_TexDim == TextureDimension.kTexDim2DArray:
            sb.append("2DArray")
        elif m_Prop.m_DefTexture.m_TexDim == TextureDimension.kTexDimCubeArray:
            sb.append("CubeArray")

    sb.append(") = ")

    if m_Prop.m_Type in [SerializedPropertyType.kColor, SerializedPropertyType.kVector]:
        sb.append(
            "({0:g},{1:g},{2:g},{3:g})".format(
                m_Prop.m_DefValue_0_,
                m_Prop.m_DefValue_1_,
                m_Prop.m_DefValue_2_,
                m_Prop.m_DefValue_3_,
            )
        )
    elif m_Prop.m_Type in [
        SerializedPropertyType.kFloat,
        SerializedPropertyType.kRange,
    ]:
        sb.append(m_Prop.m_DefValue_0_)
    elif m_Prop.m_Type == SerializedPropertyType.kTexture:
        sb.append('"{0}" {{ }}'.format(m_Prop.m_DefTexture.m_DefaultName))
    else:
        raise ValueError(m_Prop.m_Type)

    sb.append("\n")
    return "".join(map(str, sb))


def CheckGpuProgramUsable(platform: int, programType: int) -> bool:
    if platform == ShaderCompilerPlatform.kShaderCompPlatformGL:
        return programType == ShaderGpuProgramType.kShaderGpuProgramGLLegacy
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformD3D9:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramDX9VertexSM20
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX9VertexSM30
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX9PixelSM20
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX9PixelSM30
        )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformXbox360:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS
        )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformD3D11:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramDX11VertexSM40
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11VertexSM50
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11PixelSM40
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11PixelSM50
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11GeometrySM40
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11GeometrySM50
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11HullSM50
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX11DomainSM50
        )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformGLES20:
        return programType == ShaderGpuProgramType.kShaderGpuProgramGLES
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformNaCl:
        raise NotImplementedError
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformFlash:
        raise NotImplementedError
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformD3D11_9x:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramDX10Level9Vertex
            or programType == ShaderGpuProgramType.kShaderGpuProgramDX10Level9Pixel
        )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformGLES3Plus:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramGLES31AEP
            or programType == ShaderGpuProgramType.kShaderGpuProgramGLES31
            or programType == ShaderGpuProgramType.kShaderGpuProgramGLES3
        )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPSP2:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS
        )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPS4:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS
        )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformXboxOne:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS
        )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPSM:
        raise NotImplementedError
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformMetal:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramMetalVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramMetalVS
        )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformOpenGLCore:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramGLCore32
            or programType == ShaderGpuProgramType.kShaderGpuProgramGLCore41
            or programType == ShaderGpuProgramType.kShaderGpuProgramGLCore43
        )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformN3DS:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS
        )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformWiiU:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS
        )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformVulkan:
        return programType == ShaderGpuProgramType.kShaderGpuProgramSPIRV
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformSwitch:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS
        )
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformXboxOneD3D12:
        return (
            programType == ShaderGpuProgramType.kShaderGpuProgramConsoleVS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleFS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleHS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleDS
            or programType == ShaderGpuProgramType.kShaderGpuProgramConsoleGS
        )
    else:
        raise NotImplementedError


def GetPlatformString(platform: int):
    if platform == ShaderCompilerPlatform.kShaderCompPlatformGL:
        return "openGL"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformD3D9:
        return "d3d9"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformXbox360:
        return "xbox360"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPS3:
        return "ps3"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformD3D11:
        return "d3d11"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformGLES20:
        return "gles"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformNaCl:
        return "glesdesktop"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformFlash:
        return "flash"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformD3D11_9x:
        return "d3d11_9x"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformGLES3Plus:
        return "gles3"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPSP2:
        return "psp2"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPS4:
        return "ps4"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformXboxOne:
        return "xboxone"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformPSM:
        return "psm"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformMetal:
        return "metal"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformOpenGLCore:
        return "glcore"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformN3DS:
        return "n3ds"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformWiiU:
        return "wiiu"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformVulkan:
        return "vulkan"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformSwitch:
        return "switch"
    elif platform == ShaderCompilerPlatform.kShaderCompPlatformXboxOneD3D12:
        return "xboxone+_d3d12"
    else:
        return "unknown"


class ShaderProgram:
    m_SubPrograms: List[ShaderSubProgram]

    def __init__(self, reader: EndianBinaryReader, version: Tuple[int, int, int, int]):
        subProgramCapacity = reader.read_int()
        self.m_SubPrograms = [None] * subProgramCapacity

        if version >= (2019, 3):  # 2019.3 and up
            entrySize = 12
        else:
            entrySize = 8

        for i in range(subProgramCapacity):
            reader.Position = 4 + i * entrySize
            offset = reader.read_int()
            reader.Position = offset
            self.m_SubPrograms[i] = ShaderSubProgram(reader)

    def Export(self, shader: str) -> str:
        shader = re.sub(
            r"GpuProgramIndex (.+)",
            lambda math: self.m_SubPrograms[int(math.group(1))].Export(),
            shader,
        )

        return shader


class ShaderSubProgram:
    m_Version: int
    m_Keywords: List[str]
    m_ProgramCode: bytes
    m_LocalKeywords: Optional[List[str]]

    def __init__(self, reader: EndianBinaryReader):
        # LoadGpuProgramFromData
        # 201509030 - Unity 5.3
        # 201510240 - Unity 5.4
        # 201608170 - Unity 5.5
        # 201609010 - Unity 5.6, 2017.1 & 2017.2
        # 201708220 - Unity 2017.3, Unity 2017.4 & Unity 2018.1
        # 201802150 - Unity 2018.2 & Unity 2018.3
        # 201806140 - Unity 2019.1~2020.1
        # 202012090 - Unity 2021.2

        self.m_Version = reader.read_int()
        self.m_ProgramType = ShaderGpuProgramType(reader.read_int())
        reader.Position += 12

        if self.m_Version >= 201608170:
            reader.Position += 4

        m_KeywordSize = reader.read_int()
        self.m_Keywords = [reader.read_aligned_string() for _ in range(m_KeywordSize)]

        if 201806140 <= self.m_Version < 202012090:
            m_LocalKeywordsSize = reader.read_int()
            self.m_LocalKeywords = [reader.read_aligned_string() for _ in range(m_LocalKeywordsSize)]
        else:
            self.m_LocalKeywords = None

        self.m_ProgramCode = reader.read_byte_array()
        reader.align_stream()

    def Export(self) -> str:
        sb = []

        if len(self.m_Keywords) > 0:
            sb.append("Keywords { ")
            for keyword in self.m_Keywords:
                sb.append('"{0}" '.format(keyword))

            sb.append("}\n")

        if (
            getattr(self, "m_LocalKeywords") is not None  # noqa: B009
            and len(self.m_LocalKeywords) > 0  # type: ignore
        ):
            sb.append("Local Keywords { ")
            for keyword in self.m_LocalKeywords:  # type: ignore
                sb.append('"{0}" '.format(keyword))

            sb.append("}\n")

        sb.append('"')

        if len(self.m_ProgramCode) > 0:
            if self.m_ProgramType in [
                ShaderGpuProgramType.kShaderGpuProgramGLLegacy,
                ShaderGpuProgramType.kShaderGpuProgramGLES31AEP,
                ShaderGpuProgramType.kShaderGpuProgramGLES31,
                ShaderGpuProgramType.kShaderGpuProgramGLES3,
                ShaderGpuProgramType.kShaderGpuProgramGLES,
                ShaderGpuProgramType.kShaderGpuProgramGLCore32,
                ShaderGpuProgramType.kShaderGpuProgramGLCore41,
                ShaderGpuProgramType.kShaderGpuProgramGLCore43,
            ]:
                sb.append(bytes(self.m_ProgramCode).decode("utf8"))
            elif self.m_ProgramType in [
                ShaderGpuProgramType.kShaderGpuProgramDX9VertexSM20,
                ShaderGpuProgramType.kShaderGpuProgramDX9VertexSM30,
                ShaderGpuProgramType.kShaderGpuProgramDX9PixelSM20,
                ShaderGpuProgramType.kShaderGpuProgramDX9PixelSM30,
            ]:
                sb.append("// shader disassembly not supported on DXBC")
            elif self.m_ProgramType in [
                ShaderGpuProgramType.kShaderGpuProgramDX10Level9Vertex,
                ShaderGpuProgramType.kShaderGpuProgramDX10Level9Pixel,
                ShaderGpuProgramType.kShaderGpuProgramDX11VertexSM40,
                ShaderGpuProgramType.kShaderGpuProgramDX11VertexSM50,
                ShaderGpuProgramType.kShaderGpuProgramDX11PixelSM40,
                ShaderGpuProgramType.kShaderGpuProgramDX11PixelSM50,
                ShaderGpuProgramType.kShaderGpuProgramDX11GeometrySM40,
                ShaderGpuProgramType.kShaderGpuProgramDX11GeometrySM50,
                ShaderGpuProgramType.kShaderGpuProgramDX11HullSM50,
                ShaderGpuProgramType.kShaderGpuProgramDX11DomainSM50,
            ]:
                sb.append("// shader disassembly not supported on DXBC")
            elif self.m_ProgramType in [
                ShaderGpuProgramType.kShaderGpuProgramMetalVS,
                ShaderGpuProgramType.kShaderGpuProgramMetalFS,
            ]:
                reader = EndianBinaryReader(self.m_ProgramCode, endian="<")
                fourCC = reader.read_u_int()
                if fourCC == 0xF00DCAFE:
                    offset = reader.read_int()
                    reader.Position = offset

                _entryName = reader.read_string_to_null()
                buff = reader.read_bytes(int(reader.Length - reader.Position))
                sb.append(bytes(buff).decode("utf8"))
            elif self.m_ProgramType == ShaderGpuProgramType.kShaderGpuProgramSPIRV:
                # TODO SpirVShaderConverter
                pass
            elif self.m_ProgramType in [
                ShaderGpuProgramType.kShaderGpuProgramConsoleVS,
                ShaderGpuProgramType.kShaderGpuProgramConsoleFS,
                ShaderGpuProgramType.kShaderGpuProgramConsoleHS,
                ShaderGpuProgramType.kShaderGpuProgramConsoleDS,
                ShaderGpuProgramType.kShaderGpuProgramConsoleGS,
            ]:
                sb.append(bytes(self.m_ProgramCode).decode("utf8"))
            else:
                sb.append("//shader disassembly not supported on {0}".format(self.m_ProgramType))

        sb.append('"')
        return "".join(sb)



from enum import IntEnum

from .TextureFormat import TextureFormat


class GraphicsFormat(IntEnum):
    NONE = 0
    R8_SRGB = 1
    R8G8_SRGB = 2
    R8G8B8_SRGB = 3
    R8G8B8A8_SRGB = 4

    R8_UNorm = 5
    R8G8_UNorm = 6
    R8G8B8_UNorm = 7
    R8G8B8A8_UNorm = 8

    R8_SNorm = 9
    R8G8_SNorm = 10
    R8G8B8_SNorm = 11
    R8G8B8A8_SNorm = 12

    R8_UInt = 13
    R8G8_UInt = 14
    R8G8B8_UInt = 15
    R8G8B8A8_UInt = 16

    R8_SInt = 17
    R8G8_SInt = 18
    R8G8B8_SInt = 19
    R8G8B8A8_SInt = 20

    R16_UNorm = 21
    R16G16_UNorm = 22
    R16G16B16_UNorm = 23
    R16G16B16A16_UNorm = 24

    R16_SNorm = 25
    R16G16_SNorm = 26
    R16G16B16_SNorm = 27
    R16G16B16A16_SNorm = 28

    R16_UInt = 29
    R16G16_UInt = 30
    R16G16B16_UInt = 31
    R16G16B16A16_UInt = 32

    R16_SInt = 33
    R16G16_SInt = 34
    R16G16B16_SInt = 35
    R16G16B16A16_SInt = 36

    R32_UInt = 37
    R32G32_UInt = 38
    R32G32B32_UInt = 39
    R32G32B32A32_UInt = 40

    R32_SInt = 41
    R32G32_SInt = 42
    R32G32B32_SInt = 43
    R32G32B32A32_SInt = 44

    R16_SFloat = 45
    R16G16_SFloat = 46
    R16G16B16_SFloat = 47
    R16G16B16A16_SFloat = 48
    R32_SFloat = 49
    R32G32_SFloat = 50
    R32G32B32_SFloat = 51
    R32G32B32A32_SFloat = 52

    B8G8R8_SRGB = 56
    B8G8R8A8_SRGB = 57
    B8G8R8_UNorm = 58
    B8G8R8A8_UNorm = 59
    B8G8R8_SNorm = 60
    B8G8R8A8_SNorm = 61
    B8G8R8_UInt = 62
    B8G8R8A8_UInt = 63
    B8G8R8_SInt = 64
    B8G8R8A8_SInt = 65

    R4G4B4A4_UNormPack16 = 66
    B4G4R4A4_UNormPack16 = 67
    R5G6B5_UNormPack16 = 68
    B5G6R5_UNormPack16 = 69
    R5G5B5A1_UNormPack16 = 70
    B5G5R5A1_UNormPack16 = 71
    A1R5G5B5_UNormPack16 = 72

    E5B9G9R9_UFloatPack32 = 73
    B10G11R11_UFloatPack32 = 74

    A2B10G10R10_UNormPack32 = 75
    A2B10G10R10_UIntPack32 = 76
    A2B10G10R10_SIntPack32 = 77
    A2R10G10B10_UNormPack32 = 78
    A2R10G10B10_UIntPack32 = 79
    A2R10G10B10_SIntPack32 = 80
    A2R10G10B10_XRSRGBPack32 = 81
    A2R10G10B10_XRUNormPack32 = 82
    R10G10B10_XRSRGBPack32 = 83
    R10G10B10_XRUNormPack32 = 84
    A10R10G10B10_XRSRGBPack32 = 85
    A10R10G10B10_XRUNormPack32 = 86

    D16_UNorm = 90
    D24_UNorm = 91
    D24_UNorm_S8_UInt = 92
    D32_SFloat = 93
    D32_SFloat_S8_UInt = 94
    S8_UInt = 95

    RGB_DXT1_SRGB = 96
    RGBA_DXT1_SRGB = 96
    RGB_DXT1_UNorm = 97
    RGBA_DXT1_UNorm = 97
    RGBA_DXT3_SRGB = 98
    RGBA_DXT3_UNorm = 99
    RGBA_DXT5_SRGB = 100
    RGBA_DXT5_UNorm = 101
    R_BC4_UNorm = 102
    R_BC4_SNorm = 103
    RG_BC5_UNorm = 104
    RG_BC5_SNorm = 105
    RGB_BC6H_UFloat = 106
    RGB_BC6H_SFloat = 107
    RGBA_BC7_SRGB = 108
    RGBA_BC7_UNorm = 109

    RGB_PVRTC_2Bpp_SRGB = 110
    RGB_PVRTC_2Bpp_UNorm = 111
    RGB_PVRTC_4Bpp_SRGB = 112
    RGB_PVRTC_4Bpp_UNorm = 113
    RGBA_PVRTC_2Bpp_SRGB = 114
    RGBA_PVRTC_2Bpp_UNorm = 115
    RGBA_PVRTC_4Bpp_SRGB = 116
    RGBA_PVRTC_4Bpp_UNorm = 117

    RGB_ETC_UNorm = 118
    RGB_ETC2_SRGB = 119
    RGB_ETC2_UNorm = 120
    RGB_A1_ETC2_SRGB = 121
    RGB_A1_ETC2_UNorm = 122
    RGBA_ETC2_SRGB = 123
    RGBA_ETC2_UNorm = 124

    R_EAC_UNorm = 125
    R_EAC_SNorm = 126
    RG_EAC_UNorm = 127
    RG_EAC_SNorm = 128

    RGBA_ASTC4X4_SRGB = 129
    RGBA_ASTC4X4_UNorm = 130
    RGBA_ASTC5X5_SRGB = 131
    RGBA_ASTC5X5_UNorm = 132
    RGBA_ASTC6X6_SRGB = 133
    RGBA_ASTC6X6_UNorm = 134
    RGBA_ASTC8X8_SRGB = 135
    RGBA_ASTC8X8_UNorm = 136
    RGBA_ASTC10X10_SRGB = 137
    RGBA_ASTC10X10_UNorm = 138
    RGBA_ASTC12X12_SRGB = 139
    RGBA_ASTC12X12_UNorm = 140

    YUV2 = 141

    RGBA_ASTC4X4_UFloat = 145
    RGBA_ASTC5X5_UFloat = 146
    RGBA_ASTC6X6_UFloat = 147
    RGBA_ASTC8X8_UFloat = 148
    RGBA_ASTC10X10_UFloat = 149
    RGBA_ASTC12X12_UFloat = 150

    D16_UNorm_S8_UInt = 151


# very experimental & untested
GRAPHICS_TO_TEXTURE_MAP = {
    GraphicsFormat.R8_SRGB: TextureFormat.R8,
    GraphicsFormat.R8G8_SRGB: TextureFormat.RG16,
    GraphicsFormat.R8G8B8_SRGB: TextureFormat.RGB24,
    GraphicsFormat.R8G8B8A8_SRGB: TextureFormat.RGBA32,
    GraphicsFormat.R8_UNorm: TextureFormat.R8,
    GraphicsFormat.R8G8_UNorm: TextureFormat.RG16,
    GraphicsFormat.R8G8B8_UNorm: TextureFormat.RGB24,
    GraphicsFormat.R8G8B8A8_UNorm: TextureFormat.RGBA32,
    GraphicsFormat.R8_SNorm: TextureFormat.R8_SIGNED,
    GraphicsFormat.R8G8_SNorm: TextureFormat.RG16_SIGNED,
    GraphicsFormat.R8G8B8_SNorm: TextureFormat.RGB24_SIGNED,
    GraphicsFormat.R8G8B8A8_SNorm: TextureFormat.RGBA32_SIGNED,
    GraphicsFormat.R8_UInt: TextureFormat.R16,
    GraphicsFormat.R8G8_UInt: TextureFormat.RG32,
    GraphicsFormat.R8G8B8_UInt: TextureFormat.RGB48,
    GraphicsFormat.R8G8B8A8_UInt: TextureFormat.RGBA64,
    GraphicsFormat.R8_SInt: TextureFormat.R16_SIGNED,
    GraphicsFormat.R8G8_SInt: TextureFormat.RG32_SIGNED,
    GraphicsFormat.R8G8B8_SInt: TextureFormat.RGB48_SIGNED,
    GraphicsFormat.R8G8B8A8_SInt: TextureFormat.RGBA64_SIGNED,
    GraphicsFormat.R16_UNorm: TextureFormat.R16,
    GraphicsFormat.R16G16_UNorm: TextureFormat.RG32,
    GraphicsFormat.R16G16B16_UNorm: TextureFormat.RGB48,
    GraphicsFormat.R16G16B16A16_UNorm: TextureFormat.RGBA64,
    GraphicsFormat.R16_SNorm: TextureFormat.R16_SIGNED,
    GraphicsFormat.R16G16_SNorm: TextureFormat.RG32_SIGNED,
    GraphicsFormat.R16G16B16_SNorm: TextureFormat.RGB48_SIGNED,
    GraphicsFormat.R16G16B16A16_SNorm: TextureFormat.RGBA64_SIGNED,
    GraphicsFormat.R16_UInt: TextureFormat.R16,
    GraphicsFormat.R16G16_UInt: TextureFormat.RG32,
    GraphicsFormat.R16G16B16_UInt: TextureFormat.RGB48,
    GraphicsFormat.R16G16B16A16_UInt: TextureFormat.RGBA64,
    GraphicsFormat.R16_SInt: TextureFormat.R16_SIGNED,
    GraphicsFormat.R16G16_SInt: TextureFormat.RG32_SIGNED,
    GraphicsFormat.R16G16B16_SInt: TextureFormat.RGB48_SIGNED,
    GraphicsFormat.R16G16B16A16_SInt: TextureFormat.RGBA64_SIGNED,
    GraphicsFormat.B8G8R8_SRGB: TextureFormat.BGR24,
    GraphicsFormat.B8G8R8A8_SRGB: TextureFormat.BGRA32,
    GraphicsFormat.B8G8R8_UNorm: TextureFormat.BGR24,
    GraphicsFormat.B8G8R8A8_UNorm: TextureFormat.BGRA32,
    # GraphicsFormat.B8G8R8_SNorm: TextureFormat.BGR24_SIGNED,
    # GraphicsFormat.B8G8R8A8_SNorm: TextureFormat.BGRA32_SIGNED,
    # GraphicsFormat.B8G8R8_UInt: TextureFormat.BGR48,
    GraphicsFormat.RGB_DXT1_SRGB: TextureFormat.DXT1,
    GraphicsFormat.RGBA_DXT1_SRGB: TextureFormat.DXT1,
    GraphicsFormat.RGB_DXT1_UNorm: TextureFormat.DXT1,
    GraphicsFormat.RGBA_DXT1_UNorm: TextureFormat.DXT1,
    GraphicsFormat.RGBA_DXT3_SRGB: TextureFormat.DXT3,
    GraphicsFormat.RGBA_DXT3_UNorm: TextureFormat.DXT3,
    GraphicsFormat.RGBA_DXT5_SRGB: TextureFormat.DXT5,
    GraphicsFormat.RGBA_DXT5_UNorm: TextureFormat.DXT5,
    GraphicsFormat.R_BC4_UNorm: TextureFormat.BC4,
    # GraphicsFormat.R_BC4_SNorm: TextureFormat.BC4_SIGNED,
    GraphicsFormat.RG_BC5_UNorm: TextureFormat.BC5,
    # GraphicsFormat.RG_BC5_SNorm: TextureFormat.BC5_SIGNED,
    GraphicsFormat.RGB_BC6H_UFloat: TextureFormat.BC6H,
    # GraphicsFormat.RGB_BC6H_SFloat: TextureFormat.BC6H_SIGNED,
    GraphicsFormat.RGBA_BC7_SRGB: TextureFormat.BC7,
    GraphicsFormat.RGBA_BC7_UNorm: TextureFormat.BC7,
    GraphicsFormat.RGB_PVRTC_2Bpp_SRGB: TextureFormat.PVRTC_RGB2,
    GraphicsFormat.RGB_PVRTC_2Bpp_UNorm: TextureFormat.PVRTC_RGB2,
    GraphicsFormat.RGB_PVRTC_4Bpp_SRGB: TextureFormat.PVRTC_RGB4,
    GraphicsFormat.RGB_PVRTC_4Bpp_UNorm: TextureFormat.PVRTC_RGB4,
    GraphicsFormat.RGBA_PVRTC_2Bpp_SRGB: TextureFormat.PVRTC_RGBA2,
    GraphicsFormat.RGBA_PVRTC_2Bpp_UNorm: TextureFormat.PVRTC_RGBA2,
    GraphicsFormat.RGBA_PVRTC_4Bpp_SRGB: TextureFormat.PVRTC_RGBA4,
    GraphicsFormat.RGBA_PVRTC_4Bpp_UNorm: TextureFormat.PVRTC_RGBA4,
    GraphicsFormat.RGB_ETC_UNorm: TextureFormat.ETC_RGB4,
    GraphicsFormat.RGB_ETC2_SRGB: TextureFormat.ETC2_RGB,
    GraphicsFormat.RGB_ETC2_UNorm: TextureFormat.ETC2_RGB,
    GraphicsFormat.RGB_A1_ETC2_SRGB: TextureFormat.ETC2_RGBA1,
    GraphicsFormat.RGB_A1_ETC2_UNorm: TextureFormat.ETC2_RGBA1,
    GraphicsFormat.RGBA_ETC2_SRGB: TextureFormat.ETC2_RGBA8,
    GraphicsFormat.RGBA_ETC2_UNorm: TextureFormat.ETC2_RGBA8,
    GraphicsFormat.R_EAC_UNorm: TextureFormat.EAC_R,
    GraphicsFormat.R_EAC_SNorm: TextureFormat.EAC_R_SIGNED,
    GraphicsFormat.RG_EAC_UNorm: TextureFormat.EAC_RG,
    GraphicsFormat.RG_EAC_SNorm: TextureFormat.EAC_RG_SIGNED,
    GraphicsFormat.RGBA_ASTC4X4_SRGB: TextureFormat.ASTC_RGBA_4x4,
    GraphicsFormat.RGBA_ASTC4X4_UNorm: TextureFormat.ASTC_RGBA_4x4,
    GraphicsFormat.RGBA_ASTC5X5_SRGB: TextureFormat.ASTC_RGBA_5x5,
    GraphicsFormat.RGBA_ASTC5X5_UNorm: TextureFormat.ASTC_RGBA_5x5,
    GraphicsFormat.RGBA_ASTC6X6_SRGB: TextureFormat.ASTC_RGBA_6x6,
    GraphicsFormat.RGBA_ASTC6X6_UNorm: TextureFormat.ASTC_RGBA_6x6,
    GraphicsFormat.RGBA_ASTC8X8_SRGB: TextureFormat.ASTC_RGBA_8x8,
    GraphicsFormat.RGBA_ASTC8X8_UNorm: TextureFormat.ASTC_RGBA_8x8,
    GraphicsFormat.RGBA_ASTC10X10_SRGB: TextureFormat.ASTC_RGBA_10x10,
    GraphicsFormat.RGBA_ASTC10X10_UNorm: TextureFormat.ASTC_RGBA_10x10,
    GraphicsFormat.RGBA_ASTC12X12_SRGB: TextureFormat.ASTC_RGBA_12x12,
    GraphicsFormat.RGBA_ASTC12X12_UNorm: TextureFormat.ASTC_RGBA_12x12,
    GraphicsFormat.YUV2: TextureFormat.YUY2,
    # GraphicsFormat.RGBA_ASTC4X4_UFloat: TextureFormat.ASTC_RGBA_4x4,
    # GraphicsFormat.RGBA_ASTC5X5_UFloat: TextureFormat.ASTC_RGBA_5x5,
    # GraphicsFormat.RGBA_ASTC6X6_UFloat: TextureFormat.ASTC_RGBA_6x6,
    # GraphicsFormat.RGBA_ASTC8X8_UFloat: TextureFormat.ASTC_RGBA_8x8,
    # GraphicsFormat.RGBA_ASTC10X10_UFloat: TextureFormat.ASTC_RGBA_10x10,
    # GraphicsFormat.RGBA_ASTC12X12_UFloat: TextureFormat.ASTC_RGBA_12x12,
}



from __future__ import annotations

from ntpath import basename
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple

from attrs import define

from .. import config
from ..enums import BuildTarget, ClassIDType, CommonString
from ..helpers.ContainerHelper import ContainerHelper
from ..helpers.TypeTreeHelper import TypeTreeNode
from ..helpers.UnityVersion import UnityVersion
from ..streams import EndianBinaryWriter
from . import BundleFile, File, ObjectReader

if TYPE_CHECKING:
    from ..classes import AssetBundle, Object
    from ..files import ObjectReader
    from ..streams.EndianBinaryReader import EndianBinaryReader


@define(slots=True)
class SerializedFileHeader:
    metadata_size: int
    file_size: int
    version: int
    data_offset: int
    endian: str
    reserved: bytes

    def __init__(self, reader: EndianBinaryReader):
        (
            self.metadata_size,
            self.file_size,
            self.version,
            self.data_offset,
        ) = reader.read_u_int_array(4)


@define(slots=True)
class LocalSerializedObjectIdentifier:  # script type
    local_serialized_file_index: int
    local_identifier_in_file: int

    def __init__(self, header: SerializedFileHeader, reader: EndianBinaryReader):
        self.local_serialized_file_index = reader.read_int()
        if header.version < 14:
            self.local_identifier_in_file = reader.read_int()
        else:
            reader.align_stream()
            self.local_identifier_in_file = reader.read_long()

    def write(self, header: SerializedFileHeader, writer: EndianBinaryWriter):
        writer.write_int(self.local_serialized_file_index)
        if header.version < 14:
            writer.write_int(self.local_identifier_in_file)
        else:
            writer.align_stream()
            writer.write_long(self.local_identifier_in_file)


@define(slots=True)
class FileIdentifier:  # external
    path: str
    temp_empty: Optional[str] = None
    guid: Optional[bytes] = None
    type: Optional[int] = None
    # enum { kNonAssetType = 0, kDeprecatedCachedAssetType = 1, kSerializedAssetType = 2, kMetaAssetType = 3 };

    @property
    def name(self):
        return basename(self.path)

    def __repr__(self):
        return f"<{self.__class__.__name__}({self.path})>"

    def __init__(self, header: SerializedFileHeader, reader: EndianBinaryReader):
        if header.version >= 6:
            self.temp_empty = reader.read_string_to_null()
        if header.version >= 5:
            self.guid = reader.read_bytes(16)
            self.type = reader.read_int()
        self.path = reader.read_string_to_null()

    def write(self, header: SerializedFileHeader, writer: EndianBinaryWriter):
        if header.version >= 6:
            assert self.temp_empty is not None
            writer.write_string_to_null(self.temp_empty)
        if header.version >= 5:
            assert self.guid is not None and self.type is not None
            writer.write_bytes(self.guid)
            writer.write_int(self.type)
        writer.write_string_to_null(self.path)


@define(slots=True, init=False)
class SerializedType:
    class_id: int
    is_stripped_type: Optional[bool] = None
    script_type_index: int = -1
    script_id: Optional[bytes] = None  # Hash128
    old_type_hash: Optional[bytes] = None  # Hash128
    node: Optional[TypeTreeNode] = None
    # ref type
    m_ClassName: Optional[str] = None
    m_NameSpace: Optional[str] = None
    m_AssemblyName: Optional[str] = None
    # 21+
    type_dependencies: Optional[Tuple[int, ...]] = None

    def __init__(
        self,
        reader: EndianBinaryReader,
        serialized_file: SerializedFile,
        is_ref_type: bool,
    ):
        version = serialized_file.header.version
        self.class_id = reader.read_int()
        self.__attrs_init__(self.class_id)

        if version >= 16:
            self.is_stripped_type = reader.read_boolean()

        if version >= 17:
            self.script_type_index = reader.read_short()

        if version >= 13:
            if (
                (is_ref_type and self.script_type_index >= 0)
                or (version < 16 and self.class_id < 0)
                or (version >= 16 and self.class_id == 114)
            ):
                self.script_id = reader.read_bytes(16)
            self.old_type_hash = reader.read_bytes(16)

        if serialized_file._enable_type_tree:
            if version >= 12 or version == 10:
                self.node = TypeTreeNode.parse_blob(reader, version)
            else:
                self.node = TypeTreeNode.parse(reader, version)

            if version >= 21:
                if is_ref_type:
                    self.m_ClassName = reader.read_string_to_null()
                    self.m_NameSpace = reader.read_string_to_null()
                    self.m_AssemblyName = reader.read_string_to_null()
                else:
                    self.type_dependencies = reader.read_int_array()

    def write(
        self,
        serialized_file: SerializedFile,
        writer: EndianBinaryWriter,
        is_ref_type: bool,
    ):
        version = serialized_file.header.version
        writer.write_int(self.class_id)

        if version >= 16:
            assert self.is_stripped_type is not None
            writer.write_boolean(self.is_stripped_type)

        if version >= 17:
            assert self.script_type_index is not None
            writer.write_short(self.script_type_index)

        if version >= 13:
            if (
                (is_ref_type and self.script_type_index >= 0)
                or (version < 16 and self.class_id < 0)
                or (version >= 16 and self.class_id == 114)
            ):
                assert self.script_id is not None
                writer.write_bytes(self.script_id)  # Hash128
            assert self.old_type_hash is not None
            writer.write_bytes(self.old_type_hash)  # Hash128

        if serialized_file._enable_type_tree:
            assert self.node is not None
            if version >= 12 or version == 10:
                self.node.dump_blob(writer, version)
            else:
                serialized_file.dump(writer, version)

            if version >= 21:
                if is_ref_type:
                    assert (
                        self.m_ClassName is not None
                        and self.m_NameSpace is not None
                        and self.m_AssemblyName is not None
                    )
                    writer.write_string_to_null(self.m_ClassName)
                    writer.write_string_to_null(self.m_NameSpace)
                    writer.write_string_to_null(self.m_AssemblyName)
                else:
                    assert self.type_dependencies is not None
                    writer.write_int_array(self.type_dependencies, True)

    @property
    def nodes(self) -> Optional[TypeTreeNode]:
        # for compatibility with old versions
        return self.node


class SerializedFile(File.File):
    reader: EndianBinaryReader
    version: UnityVersion
    unity_version: str
    target_platform: BuildTarget
    _enable_type_tree: bool
    types: List[SerializedType]
    script_types: List[LocalSerializedObjectIdentifier]
    externals: List[FileIdentifier]
    ref_types: Optional[List[SerializedType]]
    objects: Dict[int, ObjectReader]
    unknown: int
    header: SerializedFileHeader
    _m_target_platform: int
    big_id_enabled: int
    userInformation: Optional[str]
    assetbundle: Optional[AssetBundle]
    _cache: Dict[str, Object]

    @property
    def files(self):
        if self.objects:
            return self.objects
        return {}

    @files.setter
    def files(self, value):
        self.objects = value

    def __init__(self, reader: EndianBinaryReader, parent=None, name=None, **kwargs):
        super().__init__(parent=parent, name=name, **kwargs)
        self.reader = reader

        self.unity_version = "2.5.0f5"
        self.target_platform = BuildTarget.UnknownPlatform
        self._enable_type_tree = True
        self.types = []
        self.script_types = []
        self.externals = []
        self.objects = {}
        # used to speed up mass asset extraction
        # some assets refer to each other, so by keeping the result
        # of specific assets cached the extraction can be speed up by a lot.
        # used by: Sprite (Texture2D (with alpha) cached),
        self._cache = {}
        self.unknown = 0

        # ReadHeader
        header = SerializedFileHeader(reader)
        self.header = header

        if header.version >= 9:
            header.endian = ">" if reader.read_boolean() else "<"
            header.reserved = reader.read_bytes(3)
            if header.version >= 22:
                header.metadata_size = reader.read_u_int()
                header.file_size = reader.read_long()
                header.data_offset = reader.read_long()
                self.unknown = reader.read_long()  # unknown
        else:
            reader.Position = header.file_size - header.metadata_size
            header.endian = ">" if reader.read_boolean() else "<"

        reader.endian = header.endian

        if header.version >= 7:
            unity_version = reader.read_string_to_null()
            self.set_version(unity_version)

        if header.version >= 8:
            self._m_target_platform = reader.read_int()
            self.target_platform = BuildTarget(self._m_target_platform)

        if header.version >= 13:
            self._enable_type_tree = reader.read_boolean()

        # ReadTypes
        type_count = reader.read_int()
        self.types = [SerializedType(reader, self, False) for _ in range(type_count)]

        self.big_id_enabled = 0
        if 7 <= header.version < 14:
            self.big_id_enabled = reader.read_int()

        # ReadObjects
        object_count = reader.read_int()
        self.objects = {}
        for _ in range(object_count):
            obj = ObjectReader.ObjectReader(self, reader)
            self.objects[obj.path_id] = obj

        # Read Scripts
        if header.version >= 11:
            script_count = reader.read_int()
            self.script_types = [LocalSerializedObjectIdentifier(header, reader) for _ in range(script_count)]

        # Read Externals
        externals_count = reader.read_int()
        self.externals = [FileIdentifier(header, reader) for _ in range(externals_count)]

        if header.version >= 20:
            ref_type_count = reader.read_int()
            self.ref_types = [SerializedType(reader, self, True) for _ in range(ref_type_count)]

        if config.SERIALIZED_FILE_PARSE_TYPETREE is False:
            self._enable_type_tree = False

        if header.version >= 5:
            self.userInformation = reader.read_string_to_null()

        # read the asset_bundles to get the containers
        for obj in self.objects.values():
            if obj.type == ClassIDType.AssetBundle:
                self.assetbundle = obj.read()
                self._container = ContainerHelper(self.assetbundle)
                break
        else:
            self.assetbundle = None
            self._container = ContainerHelper([])

    @property
    def container(self):
        return self._container

    def load_dependencies(self, possible_dependencies: Optional[list] = None):
        """Load all external dependencies.

        Parameters
        ----------
        possible_dependencies : list
            List of possible dependencies for cases
            where the target file is not listed as external.
        """
        for file_id in self.externals:
            self.environment.load_file(file_id.path, True)

        if possible_dependencies is None:
            return

        for dependency in possible_dependencies:
            try:
                self.environment.load_file(dependency, True)
            except FileNotFoundError:
                pass

    def set_version(self, string_version: str):
        self.unity_version = string_version
        if not string_version or string_version == "0.0.0":
            # weird case, but apparently can happen?
            # check "cant read Texture2D by 2020.3.13 f1 AssetBundle #77" for details
            if isinstance(self.parent, BundleFile.BundleFile):
                string_version = self.parent.version_engine
            if not string_version or string_version == "0.0.0":
                string_version = config.get_fallback_version()
        self.version = UnityVersion.from_str(string_version)

    def get_writeable_cab(self, name: str = "CAB-UnityPy_Mod.resS"):
        """
        Creates a new cab file in the bundle that contains the given data.
        This is usefull for asset types that use resource files.
        """
        if not isinstance(self.parent, (File.BundleFile.BundleFile, File.WebFile.WebFile)):
            return None

        cab = self.parent.get_writeable_cab(name)
        cab.path = f"archive:/{self.name}/{name}"
        if not any(cab.path == x.path for x in self.externals):
            # register as external
            class FileIdentifierFake:
                pass

            file_identifier = FileIdentifierFake()
            file_identifier.__class__ = FileIdentifier
            file_identifier.temp_empty = ""
            import uuid

            file_identifier.guid = uuid.uuid1().urn[-16:].encode("ascii")
            file_identifier.path = cab.path
            file_identifier.type = 0
            self.externals.append(file_identifier)

        return cab

    def save(self, packer: Optional[str] = None) -> bytes:
        # 1. header -> has to be delayed until the very end
        # 2. data -> types, objects, scripts, ...

        # so write the data first
        header = self.header
        meta_writer = EndianBinaryWriter(endian=header.endian)
        data_writer = EndianBinaryWriter(endian=header.endian)

        if header.version >= 7:
            meta_writer.write_string_to_null(self.unity_version)

        if header.version >= 8:
            meta_writer.write_int(self._m_target_platform)

        if header.version >= 13:
            meta_writer.write_boolean(self._enable_type_tree)

        # ReadTypes
        meta_writer.write_int(len(self.types))
        for typ in self.types:
            typ.write(self, meta_writer, False)

        if 7 <= header.version < 14:
            meta_writer.write_int(self.big_id_enabled)

        # ReadObjects
        meta_writer.write_int(len(self.objects))
        for obj in self.objects.values():
            obj.write(header, meta_writer, data_writer)
            data_writer.align_stream(8)

        # Read Scripts
        if header.version >= 11:
            meta_writer.write_int(len(self.script_types))
            for script_type in self.script_types:
                script_type.write(header, meta_writer)

        # Read Externals
        meta_writer.write_int(len(self.externals))
        for external in self.externals:
            external.write(header, meta_writer)

        if header.version >= 20:
            assert self.ref_types is not None
            meta_writer.write_int(len(self.ref_types))
            for ref_type in self.ref_types:
                ref_type.write(self, meta_writer, True)

        if header.version >= 5:
            assert self.userInformation is not None
            meta_writer.write_string_to_null(self.userInformation)

        # prepare header
        writer = EndianBinaryWriter()
        header_size = 16  # 4*4
        metadata_size = meta_writer.Length
        data_size = data_writer.Length
        if header.version >= 9:
            # 1 bool + 3 reserved + extra header 4 + 3*8
            header_size += 4 if header.version < 22 else 4 + 28
            data_offset = header_size + metadata_size
            # align data_offset
            data_offset += (16 - data_offset % 16) % 16
            file_size = data_offset + data_size
            if header.version < 22:
                writer.write_u_int(metadata_size)
                writer.write_u_int(file_size)
                writer.write_u_int(header.version)
                # reader.Position = header.file_size - header.metadata_size
                # so data follows right after this header -> after 32
                writer.write_u_int(data_offset)
                writer.write_boolean(">" == header.endian)
                writer.write_bytes(header.reserved)
            else:
                # old header
                writer.write_u_int(0)
                writer.write_u_int(0)
                writer.write_u_int(header.version)
                writer.write_u_int(0)
                writer.write_boolean(">" == header.endian)
                writer.write_bytes(header.reserved)
                writer.write_u_int(metadata_size)
                writer.write_long(file_size)
                writer.write_long(data_offset)
                writer.write_long(self.unknown)

            writer.write_bytes(meta_writer.bytes)
            writer.align_stream(16)
            writer.write_bytes(data_writer.bytes)

        else:
            metadata_size += 1  # endian boolean
            file_size = header_size + metadata_size + data_size
            writer.write_u_int(metadata_size)
            writer.write_u_int(file_size)
            writer.write_u_int(header.version)
            # reader.Position = header.file_size - header.metadata_size
            # so data follows right after this header -> after 32
            writer.write_u_int(32)
            writer.write_bytes(data_writer.bytes)
            writer.write_boolean(">" == header.endian)
            writer.write_bytes(meta_writer.bytes)

        return writer.bytes


def read_string(string_buffer_reader: EndianBinaryReader, value: int) -> str:
    is_offset = (value & 0x80000000) == 0
    if is_offset:
        string_buffer_reader.Position = value
        return string_buffer_reader.read_string_to_null()

    offset = value & 0x7FFFFFFF
    return CommonString.get(offset, str(offset))



from typing import Dict, List, Optional

from UnityPy.classes.generated import SampleClip, StreamedResource

class AudioClip(SampleClip):
    m_Name: str
    m_3D: Optional[bool] = None
    m_Ambisonic: Optional[bool] = None
    m_AudioData: Optional[List[int]] = None
    m_BitsPerSample: Optional[int] = None
    m_Channels: Optional[int] = None
    m_CompressionFormat: Optional[int] = None
    m_Format: Optional[int] = None
    m_Frequency: Optional[int] = None
    m_IsTrackerFormat: Optional[bool] = None
    m_Legacy3D: Optional[bool] = None
    m_Length: Optional[float] = None
    m_LoadInBackground: Optional[bool] = None
    m_LoadType: Optional[int] = None
    m_PreloadAudioData: Optional[bool] = None
    m_Resource: Optional[StreamedResource] = None
    m_Stream: Optional[int] = None
    m_SubsoundIndex: Optional[int] = None
    m_Type: Optional[int] = None
    m_UseHardware: Optional[bool] = None

    @property
    def extension(self) -> str: ...
    @property
    def samples(self) -> Dict[str, bytes]: ...



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from ..classes.generated import SkinnedMeshRenderer
from .MeshExporter import export_mesh_obj

if TYPE_CHECKING:
    from ..classes import (
        Material,
        Mesh,
        MeshFilter,
        PPtr,
        Renderer,
        StaticBatchInfo,
        Texture2D,
    )

    class Renderer(Renderer):
        m_Materials: Optional[list[PPtr[Material]]] = None
        m_StaticBatchInfo: Optional[StaticBatchInfo] = None
        m_SubsetIndices: Optional[List[int]] = None


def get_mesh(meshR: Renderer) -> Optional[Mesh]:
    if isinstance(meshR, SkinnedMeshRenderer):
        if meshR.m_Mesh:
            return meshR.m_Mesh.read()
    else:
        m_GameObject = meshR.m_GameObject.read()
        for comp in m_GameObject.m_Component:
            if isinstance(comp, tuple):
                pptr = comp[1]
            else:
                pptr = comp.component
            if not pptr:
                continue
            obj = pptr.deref()
            if not obj:
                continue
            if obj.type.name == "MeshFilter":
                filter: MeshFilter = pptr.read()
                if filter.m_Mesh:
                    return filter.m_Mesh.read()
    return None


def export_mesh_renderer(renderer: Renderer, export_dir: str) -> None:
    env = renderer.object_reader.assets_file.environment
    env.fs.makedirs(export_dir, exist_ok=True)
    mesh = get_mesh(renderer)
    if not mesh:
        return

    firstSubMesh = 0
    if hasattr(renderer, "m_StaticBatchInfo") and renderer.m_StaticBatchInfo.subMeshCount > 0:
        firstSubMesh = renderer.m_StaticBatchInfo.firstSubMesh
    elif hasattr(renderer, "m_SubsetIndices"):
        firstSubMesh = min(renderer.m_SubsetIndices)

    materials = []
    material_names = []
    for i in range(len(mesh.m_SubMeshes)):
        mat_index = i - firstSubMesh
        if mat_index < 0 or mat_index >= len(renderer.m_Materials):
            continue
        matPtr: Optional[PPtr[Material]] = renderer.m_Materials[i - firstSubMesh]
        if matPtr:
            mat: Material = matPtr.read()
        else:
            material_names.append(None)
            continue
        materials.append(export_material(mat))
        material_names.append(mat.m_Name)
        # save material textures
        for key, texEnv in mat.m_SavedProperties.m_TexEnvs:
            if not texEnv.m_Texture:
                continue
            if not isinstance(key, str):
                # FastPropertyName
                key = key.name
            tex: Texture2D = texEnv.m_Texture.read()
            texName = f"{tex.m_Name if tex.m_Name else key}.png"
            with env.fs.open(env.fs.sep.join([export_dir, texName]), "wb") as f:
                tex.read().image.save(f)

    # save .obj
    with env.fs.open(
        env.fs.sep.join([export_dir, f"{mesh.m_Name}.obj"]),
        "wt",
        encoding="utf8",
        newline="",
    ) as f:
        f.write(export_mesh_obj(mesh, material_names))

    # save .mtl
    with env.fs.open(
        env.fs.sep.join([export_dir, f"{mesh.m_Name}.mtl"]),
        "wt",
        encoding="utf8",
        newline="",
    ) as f:
        f.write("\n".join(materials))


def export_material(mat: Material) -> str:
    """Creates a material file (.mtl) for the given material."""

    def clt(color):  # color to tuple
        return color if isinstance(color, tuple) else (color.R, color.G, color.B, color.A)

    def properties_to_dict(properties):
        return {k if isinstance(k, str) else k.name: v for k, v in properties if v is not None}

    colors = properties_to_dict(mat.m_SavedProperties.m_Colors)
    floats = properties_to_dict(mat.m_SavedProperties.m_Floats)
    texEnvs = properties_to_dict(mat.m_SavedProperties.m_TexEnvs)

    diffuse = clt(colors.get("_Color", (0.8, 0.8, 0.8, 1)))
    ambient = clt(colors.get("_SColor", (0.2, 0.2, 0.2, 1)))
    # emissive = clt(colors.get("_EmissionColor", (0, 0, 0, 1)))
    specular = clt(colors.get("_SpecularColor", (0.2, 0.2, 0.2, 1)))
    # reflection = clt(colors.get("_ReflectColor", (0, 0, 0, 1)))
    shininess = floats.get("_Shininess", 20.0)
    transparency = floats.get("_Transparency", 0.0)

    sb: List[str] = []
    sb.append(f"newmtl {mat.m_Name}")

    # Ka r g b
    # defines the ambient color of the material to be (r,g,b). The default is (0.2,0.2,0.2);
    sb.append(f"Ka {ambient[0]:.4f} {ambient[1]:.4f} {ambient[2]:.4f}")

    # Kd r g b
    # defines the diffuse color of the material to be (r,g,b). The default is (0.8,0.8,0.8);
    sb.append(f"Kd {diffuse[0]:.4f} {diffuse[1]:.4f} {diffuse[2]:.4f}")

    # Ks r g b
    # defines the specular color of the material to be (r,g,b). This color shows up in highlights.
    # The default is (1.0,1.0,1.0);
    sb.append(f"Ks {specular[0]:.4f} {specular[1]:.4f} {specular[2]:.4f}")

    #  d alpha
    # defines the non-transparency of the material to be alpha.
    # The default is 1.0 (not transparent at all). The quantities d and Tr are the opposites of each other,
    # and specifying transparency or nontransparency is simply a matter of user convenience.

    # Tr alpha
    # defines the transparency of the material to be alpha. The default is 0.0 (not transparent at all).
    # The quantities d and Tr are the opposites of each other,
    # and specifying transparency or nontransparency is simply a matter of user convenience.
    sb.append(f"Tr {transparency:.4f}")

    # Ns s
    # defines the shininess of the material to be s. The default is 0.0;
    sb.append(f"Ns {shininess:.4f}")

    # illum n
    # denotes the illumination model used by the material.
    # illum = 1 indicates a flat material with no specular highlights,
    # so the value of Ks is not used. illum = 2 denotes the presence of specular highlights,
    # and so a specification for Ks is required.

    # map_Ka filename
    # names a file containing a texture map, which should just be an ASCII dump of RGB values;
    texName = None
    tex = None
    for key, texEnv in texEnvs:
        if not texEnv.m_Texture:
            continue
        if not isinstance(key, str):
            # FastPropertyName
            key = key.name

        tex: Texture2D = texEnv.m_Texture.read()
        texName = f"{tex.m_Name if tex.m_Name else key}.png"
        if key == "_MainTex":
            sb.append(f"map_Kd {texName}")
        elif key == "_BumpMap":
            # TODO: bump is default, some use map_bump
            sb.append(f"map_bump {texName}")
            sb.append(f"bump {texName}")
        elif "Specular" in key:
            sb.append(f"map_Ks {texName}")
        elif "Normal" in key:
            # TODO: figure out the key
            pass
    ret = "\n".join(sb)
    return ret



from __future__ import annotations

from typing import TYPE_CHECKING, List

from ...enums.GraphicsFormat import GRAPHICS_TO_TEXTURE_MAP, GraphicsFormat
from ..generated import Texture2DArray

if TYPE_CHECKING:
    from PIL import Image


def _Texture2DArray_get_images(self: Texture2DArray) -> List[Image.Image]:
    from ...export import Texture2DConverter
    from ...helpers.ResourceReader import get_resource_data

    texture_format = GRAPHICS_TO_TEXTURE_MAP.get(GraphicsFormat(self.m_Format))
    if not texture_format:
        raise NotImplementedError(f"GraphicsFormat {self.m_Format} not supported yet")

    image_data = self.image_data
    if image_data is None:
        image_data = get_resource_data(
            self.m_StreamData.path,
            self.object_reader.assets_file,
            self.m_StreamData.offset,
            self.m_StreamData.size,
        )

    # calculate the number of textures in the array
    texture_size = self.m_DataSize // self.m_Depth

    return [
        Texture2DConverter.parse_image_data(
            image_data[offset : offset + texture_size],
            self.m_Width,
            self.m_Height,
            texture_format,
            self.object_reader.version,
            0,
            None,
        )
        for offset in range(0, self.m_DataSize, texture_size)
    ]


Texture2DArray.images = property(_Texture2DArray_get_images)

__all__ = ("Texture2DArray",)



from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from ..helpers.MeshHelper import MeshHandler

if TYPE_CHECKING:
    from ..classes.generated import Mesh


def export_mesh(m_Mesh: Mesh, format: str = "obj") -> str:
    if format == "obj":
        return export_mesh_obj(m_Mesh)
    raise NotImplementedError(f"Export format {format} not implemented")


def export_mesh_obj(mesh: Mesh, material_names: Optional[List[str]] = None) -> str:
    handler = MeshHandler(mesh)
    handler.process()

    m_Mesh = handler
    if m_Mesh.m_VertexCount <= 0:
        return False

    sb = [f"g {mesh.m_Name}\n"]
    if material_names:
        sb.append(f"mtllib {mesh.m_Name}.mtl\n")
    # region Vertices
    if not m_Mesh.m_Vertices:
        return False

    sb.extend(
        "v {0:.9G} {1:.9G} {2:.9G}\n".format(-pos[0], pos[1], pos[2]).replace("nan", "0") for pos in m_Mesh.m_Vertices
    )
    # endregion

    # region UV
    if m_Mesh.m_UV0:
        sb.extend("vt {0:.9G} {1:.9G}\n".format(uv[0], uv[1]).replace("nan", "0") for uv in m_Mesh.m_UV0)
    # endregion

    # region Normals
    if m_Mesh.m_Normals:
        sb.extend(
            "vn {0:.9G} {1:.9G} {2:.9G}\n".format(-n[0], n[1], n[2]).replace("nan", "0") for n in m_Mesh.m_Normals
        )
    # endregion

    # region Face
    for i, triangles in enumerate(m_Mesh.get_triangles()):
        sb.append(f"g {mesh.m_Name}_{i}\n")
        if material_names and i < len(material_names) and material_names[i]:
            sb.append(f"usemtl {material_names[i]}\n")
        sb.extend("f {0}/{0}/{0} {1}/{1}/{1} {2}/{2}/{2}\n".format(c + 1, b + 1, a + 1) for a, b, c in triangles)
    # endregion
    return "".join(sb)



from __future__ import annotations

import io
import os
from typing import List, Optional, Tuple, Union

from .. import files
from ..enums import FileType
from ..streams import EndianBinaryReader
from .CompressionHelper import BROTLI_MAGIC, GZIP_MAGIC

FileSourceType = Union[str, bytes, bytearray, io.IOBase, EndianBinaryReader]


def file_name_without_extension(file_name: str) -> str:
    return os.path.join(os.path.dirname(file_name), os.path.splitext(os.path.basename(file_name))[0])


def list_all_files(directory: str) -> List[str]:
    return [
        val
        for sublist in [
            [os.path.join(dir_path, filename) for filename in filenames]
            for (dir_path, dirnames, filenames) in os.walk(directory)
            if ".git" not in dir_path
        ]
        for val in sublist
    ]


def find_all_files(directory: str, search_str: str) -> List[str]:
    return [
        val
        for sublist in [
            [os.path.join(dir_path, filename) for filename in filenames if search_str in filename]
            for (dir_path, dirnames, filenames) in os.walk(directory)
            if ".git" not in dir_path
        ]
        for val in sublist
    ]


def check_file_type(
    input_: FileSourceType,
) -> Tuple[FileType, EndianBinaryReader]:
    if isinstance(input_, str) and os.path.isfile(input_):
        reader = EndianBinaryReader(open(input_, "rb"))
    elif isinstance(input_, EndianBinaryReader):
        reader = input_
    else:
        reader = EndianBinaryReader(input_)

    if reader.Length < 20:
        return FileType.ResourceFile, reader

    signature = reader.read_string_to_null(20)

    reader.Position = 0
    if signature in [
        "UnityWeb",
        "UnityRaw",
        "\xfa\xfa\xfa\xfa\xfa\xfa\xfa\xfa",
        "UnityFS",
    ]:
        return FileType.BundleFile, reader
    elif signature.startswith(("UnityWebData", "TuanjieWebData")):
        return FileType.WebFile, reader
    elif signature == "PK\x03\x04":
        return FileType.ZIP, reader
    else:
        if reader.Length < 128:
            return FileType.ResourceFile, reader

        magic = bytes(reader.read_bytes(2))
        reader.Position = 0
        if GZIP_MAGIC == magic:
            return FileType.WebFile, reader
        reader.Position = 0x20
        magic = bytes(reader.read_bytes(6))
        reader.Position = 0
        if BROTLI_MAGIC == magic:
            return FileType.WebFile, reader

        # check if AssetsFile
        old_endian = reader.endian
        # read as if assetsfile and check version
        # ReadHeader
        reader.Position = 0
        metadata_size = reader.read_u_int()
        file_size = reader.read_u_int()
        version = reader.read_u_int()
        data_offset = reader.read_u_int()

        if version >= 22:
            raw_endian = reader.read_u_byte()
            _endian = ">" if raw_endian else "<"
            _reserved = reader.read_bytes(3)
            metadata_size = reader.read_u_int()
            file_size = reader.read_long()
            data_offset = reader.read_long()
            _unknown = reader.read_long()  # unknown

        # reset
        reader.endian = old_endian
        reader.Position = 0
        # check info
        if any(
            (
                version < 0,
                version > 100,
                *[x < 0 or x > reader.Length for x in [file_size, metadata_size, version, data_offset]],
                file_size < metadata_size,
                file_size < data_offset,
            )
        ):
            return FileType.ResourceFile, reader
        else:
            return FileType.AssetsFile, reader


def parse_file(
    reader: EndianBinaryReader,
    parent: files.File,
    name: str,
    typ: Optional[FileType] = None,
    is_dependency: bool = False,
) -> Union[files.File, EndianBinaryReader]:
    if typ is None:
        typ, _ = check_file_type(reader)
    if typ == FileType.AssetsFile and not name.endswith(
        (
            ".resS",
            ".resource",
            ".config",
            ".xml",
            ".dat",
        )
    ):
        f = files.SerializedFile(reader, parent, name=name, is_dependency=is_dependency)
    elif typ == FileType.BundleFile:
        f = files.BundleFile(reader, parent, name=name, is_dependency=is_dependency)
    elif typ == FileType.WebFile:
        f = files.WebFile(reader, parent, name=name, is_dependency=is_dependency)
    else:
        f = reader
    return f


def find_sensitive_path(dir: str, insensitive_path: str) -> Union[str, None]:
    parts = os.path.split(insensitive_path.strip(os.path.sep))

    sensitive_path = dir
    for part in parts:
        part_lower = part.lower()
        part = next(
            (name for name in os.listdir(sensitive_path) if name.lower() == part_lower),
            None,
        )
        if part is None:
            return None
        sensitive_path = os.path.join(sensitive_path, part)

    return sensitive_path



from __future__ import annotations

from typing import List, Union

from ...enums import ClassIDType
from ..generated import Component, GameObject, PPtr


def _GameObject_Components(self) -> List[PPtr[Component]]:
    if self.m_Component is None:
        return []
    if isinstance(self.m_Component[0], tuple):
        return [c.m_GameObject for i, c in self.m_Component]
    else:
        return [c.component for c in self.m_Component]


def _GameObject_GetComponent(self, type: ClassIDType) -> Union[PPtr[Component], None]:
    for component in self.m_Components:
        if component.type == type:
            return component
    return None


GameObject.m_Components = property(_GameObject_Components)
GameObject.m_Animator = property(lambda self: _GameObject_GetComponent(self, ClassIDType.Animator))
GameObject.m_Animation = property(lambda self: _GameObject_GetComponent(self, ClassIDType.Animation))
GameObject.m_Transform = property(lambda self: _GameObject_GetComponent(self, ClassIDType.Transform))
GameObject.m_MeshRenderer = property(lambda self: _GameObject_GetComponent(self, ClassIDType.MeshRenderer))
GameObject.m_SkinnedMeshRenderer = property(
    lambda self: _GameObject_GetComponent(self, ClassIDType.SkinnedMeshRenderer)
)
GameObject.m_MeshFilter = property(lambda self: _GameObject_GetComponent(self, ClassIDType.MeshFilter))
